# The MIT License (MIT)
# Copyright (c) 2020 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from abc import abstractmethod, ABC
from typing import Iterator, Tuple, Any, Optional, \
    List, Type, Dict, Union, Container

from xcube.constants import EXTENSION_POINT_DATA_STORES
from xcube.util.extension import Extension
from xcube.util.extension import ExtensionPredicate
from xcube.util.extension import ExtensionRegistry
from xcube.util.jsonschema import JsonObjectSchema
from xcube.util.plugin import get_extension_registry
from .accessor import DataOpener
from .accessor import DataWriter
from .assertions import assert_valid_params
from .datatype import DataTypeLike
from .descriptor import DataDescriptor
from .error import DataStoreError
from .search import DataSearcher


#######################################################
# Data store instantiation and registry query
#######################################################


def new_data_store(data_store_id: str,
                   extension_registry: Optional[ExtensionRegistry] = None,
                   **data_store_params) \
        -> Union['DataStore', 'MutableDataStore']:
    """
    Create a new data store instance for given
    *data_store_id* and *data_store_params*.

    :param data_store_id: A data store identifier.
    :param extension_registry: Optional extension registry.
        If not given, the global extension registry will be used.
    :param data_store_params: Data store specific parameters.
    :return: A new data store instance
    """
    data_store_class = get_data_store_class(
        data_store_id,
        extension_registry=extension_registry
    )
    data_store_params_schema = data_store_class.get_data_store_params_schema()
    assert_valid_params(data_store_params,
                        name='data_store_params',
                        schema=data_store_params_schema)
    # noinspection PyArgumentList
    return data_store_class(**data_store_params)


def get_data_store_class(
        data_store_id: str,
        extension_registry: Optional[ExtensionRegistry] = None
) -> Union[Type['DataStore'], Type['MutableDataStore']]:
    """
    Get the class for the data store identified by *data_store_id*.

    :param data_store_id: A data store identifier.
    :param extension_registry: Optional extension registry.
        If not given, the global extension registry will be used.
    :return: The class for the data store.
    """
    extension_registry = extension_registry or get_extension_registry()
    if not extension_registry.has_extension(EXTENSION_POINT_DATA_STORES,
                                            data_store_id):
        raise DataStoreError(f'Unknown data store "{data_store_id}"'
                             f' (may be due to missing xcube plugin)')
    return extension_registry.get_component(EXTENSION_POINT_DATA_STORES,
                                            data_store_id)


def get_data_store_params_schema(
        data_store_id: str,
        extension_registry: Optional[ExtensionRegistry] = None
) -> JsonObjectSchema:
    """
    Get the JSON schema for instantiating a new data store
    identified by *data_store_id*.

    :param data_store_id: A data store identifier.
    :param extension_registry: Optional extension registry.
        If not given, the global extension registry will be used.
    :return: The JSON schema for the data store's parameters.
    """
    data_store_class = get_data_store_class(
        data_store_id,
        extension_registry=extension_registry
    )
    return data_store_class.get_data_store_params_schema()


def find_data_store_extensions(
        predicate: ExtensionPredicate = None,
        extension_registry: Optional[ExtensionRegistry] = None
) -> List[Extension]:
    """
    Find data store extensions using the optional filter
    function *predicate*.

    :param predicate: An optional filter function.
    :param extension_registry: Optional extension registry.
        If not given, the global extension registry will be used.
    :return: List of data store extensions.
    """
    extension_registry = extension_registry or get_extension_registry()
    return extension_registry.find_extensions(EXTENSION_POINT_DATA_STORES,
                                              predicate=predicate)


#######################################################
# Classes
#######################################################

class DataStore(DataOpener, DataSearcher, ABC):
    """
    A data store represents a collection of data resources that
    can be enumerated, queried, and opened in order to obtain
    in-memory representations of the data. The same data resource may be
    made available using different data types. Therefore, many methods
    allow specifying a *data_type* parameter.

    A store implementation may use any existing openers/writers,
    or define its own, or not use any openers/writers at all.

    Store implementers should follow the conventions outlined in
    https://xcube.readthedocs.io/en/latest/storeconv.html .

    The :class:DataStore is an abstract base class that both read-only and
    mutable data stores must implement.
    """

    @classmethod
    def get_data_store_params_schema(cls) -> JsonObjectSchema:
        """
        Get descriptions of parameters that must or can be used to
        instantiate a new DataStore object.
        Parameters are named and described by the properties of the
        returned JSON object schema.
        The default implementation returns JSON object schema that
        can have any properties.
        """
        return JsonObjectSchema()

    @classmethod
    @abstractmethod
    def get_data_types(cls) -> Tuple[str, ...]:
        """
        Get alias names for all data types supported by this store.
        The first entry in the tuple represents this store's
        default data type.

        :return: The tuple of supported data types.
        """

    @abstractmethod
    def get_data_types_for_data(self, data_id: str) -> Tuple[str, ...]:
        """
        Get alias names for of data types that are supported
        by this store for the given *data_id*.

        :param data_id: An identifier of data that is provided by this store
        :return: A tuple of data types that apply to the given *data_id*.
        :raise DataStoreError: If an error occurs.
        """

    @abstractmethod
    def get_data_ids(self,
                     data_type: DataTypeLike = None,
                     include_attrs: Container[str] = None) -> \
            Union[Iterator[str], Iterator[Tuple[str, Dict[str, Any]]]]:
        """
        Get an iterator over the data resource identifiers for the
        given type *data_type*. If *data_type* is omitted, all data
        resource identifiers are returned.

        If a store implementation supports only a single data type,
        it should verify that *data_type* is either None or
        compatible with the supported data type.

        If *include_attrs* is provided, it must be a sequence of names
        of metadata attributes. The store will then return extra metadata
        for each returned data resource identifier according to the
        names of the metadata attributes as tuples (*data_id*, *attrs*).

        Hence, the type of the returned iterator items depends on the
        value of *include_attrs*:

        - If *include_attrs* is None (the default), the method returns
          an iterator of dataset identifiers *data_id* of type `str`.
        - If *include_attrs* is a sequence of attribute names, even an
          empty one, the method returns an iterator of tuples
          (*data_id*, *attrs*) of type `Tuple[str, Dict]`, where *attrs*
          is a dictionary filled according to the names in *include_attrs*.
          If a store cannot provide a given attribute, it should simply
          ignore it. This may even yield to an empty dictionary for a given
          *data_id*.

        The individual attributes do not have to exist in the dataset's
        metadata, they may also be generated on-the-fly.
        An example for a generic attribute name is "title".
        A store should try to resolve ``include_attrs=["title"]``
        by returning items such as
        ``("ESACCI-L4_GHRSST-SSTdepth-OSTIA-GLOB_CDR2.1-v02.0-fv01.0.zarr",
        {"title": "Level-4 GHRSST Analysed Sea Surface Temperature"})``.

        :param data_type: If given, only data identifiers that are
            available as this type are returned.
            If this is omitted, all available data identifiers are returned.
        :param include_attrs: A sequence of names of attributes to
            be returned for each dataset identifier.
            If given, the store will attempt to provide the set of
            requested dataset attributes in addition to the data ids.
            (added in xcube 0.8.0)
        :return: An iterator over the identifiers and titles of data
            resources provided by this data store.
        :raise DataStoreError: If an error occurs.
        """

    def list_data_ids(self,
                      data_type: DataTypeLike = None,
                      include_attrs: Container[str] = None) -> \
            Union[List[str], List[Tuple[str, Dict[str, Any]]]]:
        """
        Convenience version of `get_data_ids()` that returns a list rather
        than an iterator.

        :param data_type: If given, only data identifiers that are
            available as this type are returned.
            If this is omitted, all available data identifiers are returned.
        :param include_attrs: A sequence of names of attributes to
            be returned for each dataset identifier.
            If given, the store will attempt to provide the set of
            requested dataset attributes in addition to the data ids.
            (added in xcube 0.8.0)
        :return: A list comprising the identifiers and titles of data
            resources provided by this data store.
        :raise DataStoreError: If an error occurs.
        """
        return list(self.get_data_ids(data_type=data_type,
                                      include_attrs=include_attrs))

    @abstractmethod
    def has_data(self,
                 data_id: str,
                 data_type: DataTypeLike = None) -> bool:
        """
        Check if the data resource given by *data_id* is
        available in this store.

        :param data_id: A data identifier
        :param data_type: An optional data type.
            If given, it will also be checked whether the data
            is available as the specified type.
            May be given as type alias name, as a type,
            or as a :class:DataType instance.
        :return: True, if the data resource is available in this store,
            False otherwise.
        """

    @abstractmethod
    def describe_data(self,
                      data_id: str,
                      data_type: DataTypeLike = None) -> DataDescriptor:
        """
        Get the descriptor for the data resource given by *data_id*.

        Raises a :class:DataStoreError if *data_id* does not
        exist in this store or the data is not available as the
        specified *data_type*.

        :param data_id: An identifier of data provided by this store
        :param data_type: If given, the descriptor of the data will
            describe the data as specified by the data type.
            May be given as type alias name, as a type,
            or as a :class:DataType instance.
        :return a data-type specific data descriptor
        :raise DataStoreError: If an error occurs.
        """

    @abstractmethod
    def get_data_opener_ids(self,
                            data_id: str = None,
                            data_type: DataTypeLike = None) \
            -> Tuple[str, ...]:
        """
        Get identifiers of data openers that can be used to open data
        resources from this store.

        If *data_id* is given, data accessors are restricted to the ones
        that can open the identified data resource.
        Raises if *data_id* does not exist in this store.

        If *data_type* is given, only openers that are compatible with
        this data type are returned.

        If a store implementation supports only a single data type,
        it should verify that *data_type* is either None or equal to
        that single data type.

        :param data_id: An optional data resource identifier that is
            known to exist in this data store.
        :param data_type: An optional data type that is known to be
            supported by this data store.
            May be given as type alias name, as a type,
            or as a :class:DataType instance.
        :return: A tuple of identifiers of data openers that can be
            used to open data resources.
        :raise DataStoreError: If an error occurs.
        """

    @abstractmethod
    def get_open_data_params_schema(self,
                                    data_id: str = None,
                                    opener_id: str = None) \
            -> JsonObjectSchema:
        """
        Get the schema for the parameters passed as *open_params* to
        :meth:open_data(data_id, open_params).

        If *data_id* is given, the returned schema will be tailored
        to the constraints implied by the identified data resource.
        Some openers might not support this, therefore *data_id* is optional,
        and if it is omitted, the returned schema will be less restrictive.
        If given, the method raises if *data_id* does not exist in this store.

        If *opener_id* is given, the returned schema will be tailored to
        the constraints implied by the identified opener. Some openers
        might not support this, therefore *opener_id* is optional, and if
        it is omitted, the returned schema will be less restrictive.

        For maximum compatibility of stores, it is strongly encouraged to
        apply the following conventions on parameter names, types,
        and their interpretation.

        Let P be the value of an optional, data constraining open parameter,
        then it should be interpreted as follows:

          * _if P is None_ means, parameter not given,
            hence no constraint applies, hence no additional restrictions
            on requested data.
          * _if not P_ means, we exclude data that would be
            included by default.
          * _else_, the given constraint applies.

        Given here are names, types, and descriptions of common,
        constraining open parameters for gridded datasets.
        Note, whether any of these is optional or mandatory depends
        on the individual data store. A store may also
        define other open parameters or support only a subset of the
        following. Note all parameters may be optional,
        the Python-types given here refer to _given_, non-Null parameters:

          * ``variable_names: List[str]``: Included data variables.
            Available coordinate variables will be auto-included for
            any dimension of the data variables.
          * ``bbox: Tuple[float, float, float, float]``: Spatial coverage
            as xmin, ymin, xmax, ymax.
          * ``crs: str``: Spatial CRS, e.g. "EPSG:4326" or OGC CRS URI.
          * ``spatial_res: float``: Spatial resolution in
            coordinates of the spatial CRS.
          * ``time_range: Tuple[Optional[str], Optional[str]]``:
            Time range interval in UTC date/time units using ISO format.
            Start or end time may be missing which means everything until
            available start or end time.
          * ``time_period: str`: Pandas-compatible period/frequency
            string, e.g. "8D", "2W".

        E.g. applied to an optional `variable_names` parameter, this means

          * `variable_names is None` - include all data variables
          * `variable_names == []` - do not include data variables
            (schema only)
          * `variable_names == ["<var_1>", "<var_2>", ...]` only
            include data variables named "<var_1>", "<var_2>", ...

        :param data_id: An optional data identifier that is known
            to exist in this data store.
        :param opener_id: An optional data opener identifier.
        :return: The schema for the parameters in *open_params*.
        :raise DataStoreError: If an error occurs.
        """

    @abstractmethod
    def open_data(self,
                  data_id: str,
                  opener_id: str = None,
                  **open_params) -> Any:
        """
        Open the data given by the data resource identifier *data_id*
        using the supplied *open_params*.

        The data type of the return value depends on the data opener
        used to open the data resource.

        If *opener_id* is given, the identified data opener will be used
        to open the data resource and *open_params* must comply with the
        schema of the opener's parameters. Note that some store
        implementations may not support using different openers or just
        support a single one.

        Raises if *data_id* does not exist in this store.

        :param data_id: The data identifier that is known to exist
            in this data store.
        :param opener_id: An optional data opener identifier.
        :param open_params: Opener-specific parameters.
        :return: An in-memory representation of the data resources
            identified by *data_id* and *open_params*.
        :raise DataStoreError: If an error occurs.
        """


class MutableDataStore(DataStore, DataWriter, ABC):
    """
    A mutable data store is a data store that also allows for adding,
    updating, and removing data resources.

    MutableDataStore is an abstract base class that any mutable data
    store must implement.
    """

    @abstractmethod
    def get_data_writer_ids(self,
                            data_type: DataTypeLike = None) \
            -> Tuple[str, ...]:
        """
        Get identifiers of data writers that can be used to write data
        resources to this store.

        If *data_type* is given, only writers that support this
        data type are returned.

        If a store implementation supports only a single data type,
        it should verify that *data_type* is either None or equal to
        that single data type.

        :param data_type: An optional data type specifier that is
            known to be supported by this data store.
            May be given as type alias name, as a type,
            or as a :class:DataType instance.
        :return: A tuple of identifiers of data writers that can
            be used to write data resources.
        :raise DataStoreError: If an error occurs.
        """

    @abstractmethod
    def get_write_data_params_schema(self,
                                     writer_id: str = None) \
            -> JsonObjectSchema:
        """
        Get the schema for the parameters passed as *write_params* to
        :meth:write_data(data, data_id, open_params).

        If *writer_id* is given, the returned schema will be tailored
        to the constraints implied by the identified writer.
        Some writers might not support this, therefore *writer_id*
        is optional, and if it is omitted, the returned schema will
        be less restrictive.

        Given here is a pseudo-code implementation for stores that support
        multiple writers and where the store has common parameters with
        the writer:

            store_params_schema = self.get_data_store_params_schema()
            writer_params_schema = get_writer(writer_id).get_write_data_params_schema()
            return subtract_param_schemas(writer_params_schema, store_params_schema)

        :param writer_id: An optional data writer identifier.
        :return: The schema for the parameters in *write_params*.
        :raise DataStoreError: If an error occurs.
        """

    @abstractmethod
    def write_data(self,
                   data: Any,
                   data_id: str = None,
                   writer_id: str = None,
                   replace: bool = False,
                   **write_params) -> str:
        """
        Write a data in-memory instance using the supplied *data_id*
         and *write_params*.

        If data identifier *data_id* is not given, a writer-specific default
        will be generated, used, and returned.

        If *writer_id* is given, the identified data writer will be used to
        write the data resource and *write_params* must comply with the
        schema of writers's parameters. Note that some store
        implementations may not support using different writers or
        just support a single one.

        Given here is a pseudo-code implementation for stores that support
        multiple writers:

            writer_id = writer_id or self.gen_data_id()
            path = self.resolve_data_id_to_path(data_id)
            write_params = add_params(self.get_data_store_params(), write_params)
            get_writer(writer_id).write_data(data, path, **write_params)
            self.register_data(data_id, data)

        Raises if *data_id* does not exist in this store.

        :param data: The data in-memory instance to be written.
        :param data_id: An optional data identifier that is known to
            be unique in this data store.
        :param writer_id: An optional data writer identifier.
        :param replace: Whether to replace an existing data resource.
        :param write_params: Writer-specific parameters.
        :return: The data identifier used to write the data.
        :raise DataStoreError: If an error occurs.
        """

    @abstractmethod
    def delete_data(self, data_id: str, **delete_params):
        """
        Delete the data resource identified by *data_id*.

        Typically, an implementation would delete the data resource
        from the physical storage and also remove any registered metadata
        from an associated database.

        Raises if *data_id* does not exist in this store.

        :param data_id: An data identifier that is known to exist
            in this data store.
        :raise DataStoreError: If an error occurs.
        """

    @abstractmethod
    def register_data(self, data_id: str, data: Any):
        """
        Register the in-memory representation of a data resource *data*
        using the given data resource identifier *data_id*.

        This method can be used to register data resources that are
        already physically stored in the data store, but are not yet
        searchable or otherwise accessible by the given *data_id*.

        Typically, an implementation would extract metadata from
        *data* and store it in a store-specific database.
        An implementation should just store the metadata of *data*.
        It should not write *data*.

        :param data_id: A data resource identifier that is known
            to be unique in this data store.
        :param data: An in-memory representation of a data resource.
        :raise DataStoreError: If an error occurs.
        """

    @abstractmethod
    def deregister_data(self, data_id: str):
        """
        De-register a data resource identified by *data_id* from
        this data store.

        This method can be used to de-register data resources so it
        will be no longer searchable or otherwise accessible by
        the given *data_id*.

        Typically, an implementation would extract metadata from
        *data* and store it in a store-specific database.
        An implementation should only remove a data resource's metadata.
        It should not delete *data* from its physical storage space.

        Raises if *data_id* does not exist in this store.

        :param data_id: A data resource identifier that is
            known to exist in this data store.
        :raise DataStoreError: If an error occurs.
        """



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import warnings
from collections.abc import MutableMapping
from typing import Optional, Dict, Any, Hashable, Union, Set, List, Tuple

import numpy as np
import pyproj
import xarray as xr
import zarr
import zarr.convenience

from xcube.core.schema import get_dataset_chunks
from xcube.util.assertions import assert_instance
from .base import CRS_WGS84

class GridCoords:
    """
    Grid coordinates comprising x and y of
    type xarray.DataArray.
    """

    def __init__(self):
        self.x: Optional[xr.DataArray] = None
        self.y: Optional[xr.DataArray] = None


class GridMappingProxy:
    """
    Grid mapping comprising *crs* of type pyproj.crs.CRS,
    grid coordinates, an optional name, coordinates, and a
    tile size (= spatial chunk sizes).
    """

    def __init__(self,
                 crs: Optional[pyproj.crs.CRS] = None,
                 name: Optional[str] = None,
                 coords: Optional[GridCoords] = None,
                 tile_size: Optional[Tuple[int, int]] = None):
        self.crs = crs
        self.name = name
        self.coords = coords
        self.tile_size = tile_size


def get_dataset_grid_mapping_proxies(
        dataset: xr.Dataset,
        *,
        missing_latitude_longitude_crs: pyproj.crs.CRS = None,
        missing_rotated_latitude_longitude_crs: pyproj.crs.CRS = None,
        missing_projected_crs: pyproj.crs.CRS = None,
        emit_warnings: bool = False
) -> Dict[Union[Hashable, None], GridMappingProxy]:
    """
    Find grid mappings encoded as described in the CF conventions
    [Horizontal Coordinate Reference Systems, Grid Mappings, and Projections]
    (http://cfconventions.org/Data/cf-conventions/cf-conventions-1.8/cf-conventions.html#grid-mappings-and-projections).

    :param dataset:
    :param missing_latitude_longitude_crs:
    :param missing_rotated_latitude_longitude_crs:
    :param missing_projected_crs:
    :param emit_warnings:
    :return:
    """
    grid_mapping_proxies: Dict[Union[Hashable, None],
                               GridMappingProxy] = dict()

    # Find any grid mapping variables by CF 'grid_mapping' attribute
    #
    for var_name, var in dataset.variables.items():
        grid_mapping_var_name = var.attrs.get('grid_mapping')
        if grid_mapping_var_name \
                and grid_mapping_var_name not in grid_mapping_proxies \
                and grid_mapping_var_name in dataset:
            grid_mapping_var = dataset[grid_mapping_var_name]
            gmp = _parse_crs_from_attrs(grid_mapping_var.attrs)
            grid_mapping_proxies[grid_mapping_var_name] = gmp

    # If no grid mapping variables found,
    # try if CRS is encoded in some variable's attributes
    #
    if not grid_mapping_proxies:
        for var_name, var in dataset.variables.items():
            gmp = _parse_crs_from_attrs(var.attrs)
            if gmp is not None:
                grid_mapping_proxies[var_name] = gmp
                break

    # If no grid mapping variables found,
    # try if CRS is encoded in dataset attributes
    #
    if not grid_mapping_proxies:
        gmp = _parse_crs_from_attrs(dataset.attrs)
        if gmp is not None:
            grid_mapping_proxies[None] = gmp

    # Find coordinate variables.
    #

    latitude_longitude_coords = GridCoords()
    rotated_latitude_longitude_coords = GridCoords()
    projected_coords = GridCoords()

    potential_coord_vars = _find_potential_coord_vars(dataset)

    # Find coordinate variables that use a CF standard_name.
    #
    coords_standard_names = (
        (latitude_longitude_coords,
         'longitude', 'latitude'),
        (rotated_latitude_longitude_coords,
         'grid_longitude', 'grid_latitude'),
        (projected_coords,
         'projection_x_coordinate', 'projection_y_coordinate')
    )
    for var_name in potential_coord_vars:
        var = dataset[var_name]
        if var.ndim not in (1, 2):
            continue
        standard_name = var.attrs.get('standard_name')
        for coords, x_name, y_name in coords_standard_names:
            if coords.x is None and standard_name == x_name:
                coords.x = var
            if coords.y is None and standard_name == y_name:
                coords.y = var

    # Find coordinate variables by common naming convention.
    #
    coords_var_names = (
        (latitude_longitude_coords,
         ('lon', 'longitude'),
         ('lat', 'latitude')),
        (rotated_latitude_longitude_coords,
         ('rlon', 'rlongitude'),
         ('rlat', 'rlatitude')),
        (projected_coords,
         ('x', 'xc', 'transformed_x'),
         ('y', 'yc', 'transformed_y'))
    )
    for var_name in potential_coord_vars:
        var = dataset[var_name]
        if var.ndim not in (1, 2):
            continue
        for coords, x_names, y_names in coords_var_names:
            if coords.x is None and var_name in x_names:
                coords.x = var
            if coords.y is None and var_name in y_names:
                coords.y = var

    # Assign found coordinates to grid mappings
    #
    for gmp in grid_mapping_proxies.values():
        if gmp.name == 'latitude_longitude':
            gmp.coords = latitude_longitude_coords
        elif gmp.name == 'rotated_latitude_longitude':
            gmp.coords = rotated_latitude_longitude_coords
        else:
            gmp.coords = projected_coords

    _complement_grid_mapping_coords(latitude_longitude_coords,
                                    'latitude_longitude',
                                    missing_latitude_longitude_crs
                                    or CRS_WGS84,
                                    grid_mapping_proxies)
    _complement_grid_mapping_coords(rotated_latitude_longitude_coords,
                                    'rotated_latitude_longitude',
                                    missing_rotated_latitude_longitude_crs,
                                    grid_mapping_proxies)
    _complement_grid_mapping_coords(projected_coords,
                                    None,
                                    missing_projected_crs,
                                    grid_mapping_proxies)

    # Collect complete grid mappings
    complete_grid_mappings = dict()
    for var_name, gmp in grid_mapping_proxies.items():
        if gmp.coords is not None \
                and gmp.coords.x is not None \
                and gmp.coords.y is not None \
                and gmp.coords.x.size >= 2 \
                and gmp.coords.y.size >= 2 \
                and gmp.coords.x.ndim == gmp.coords.y.ndim:
            if gmp.coords.x.ndim == 1:
                gmp.tile_size = _find_dataset_tile_size(
                    dataset,
                    gmp.coords.x.dims[0],
                    gmp.coords.y.dims[0]
                )
                complete_grid_mappings[var_name] = gmp
            elif gmp.coords.x.ndim == 2 \
                    and gmp.coords.x.dims == gmp.coords.y.dims:
                gmp.tile_size = _find_dataset_tile_size(
                    dataset,
                    gmp.coords.x.dims[1],
                    gmp.coords.x.dims[0]
                )
                complete_grid_mappings[var_name] = gmp
        elif emit_warnings:
            warnings.warn(f'CRS "{gmp.name}": '
                          f'missing x- and/or y-coordinates '
                          f'(grid mapping variable "{var_name}": '
                          f'grid_mapping_name="{gmp.name}")')

    return complete_grid_mappings


def _parse_crs_from_attrs(attrs: Dict[Hashable, Any]) \
        -> Optional[GridMappingProxy]:
    # noinspection PyBroadException
    try:
        crs = pyproj.crs.CRS.from_cf(attrs)
    except pyproj.crs.CRSError:
        return None
    return GridMappingProxy(crs=crs,
                            name=attrs.get('grid_mapping_name'))


def _complement_grid_mapping_coords(
        coords: GridCoords,
        grid_mapping_name: Optional[str],
        missing_crs: Optional[pyproj.crs.CRS],
        grid_mappings: Dict[Optional[str], GridMappingProxy]
):
    if coords.x is not None or coords.y is not None:
        grid_mapping = next((grid_mapping
                             for grid_mapping in grid_mappings.values()
                             if grid_mapping_name is None
                             or grid_mapping_name == grid_mapping.name),
                            None)
        if grid_mapping is None and missing_crs is not None:
            grid_mapping = GridMappingProxy(crs=missing_crs,
                                            name=grid_mapping_name)
            grid_mappings[None] = grid_mapping

        if grid_mapping is not None:
            if grid_mapping.coords is None:
                grid_mapping.coords = coords
            # Edge case from GeoTIFF with CRS-84 with 1D
            # coordinates named "x" and "y" as read by rioxarray
            if grid_mapping.coords.x is None:
                grid_mapping.coords.x = coords.x
            if grid_mapping.coords.y is None:
                grid_mapping.coords.y = coords.y


def _find_potential_coord_vars(dataset: xr.Dataset) -> List[Hashable]:
    """
    Find potential coordinate variables.

    We need this function as we can not rely on xarray.Dataset.coords,
    because 2D coordinate arrays are most likely not indicated as such
    in many datasets.
    """

    # Collect bounds variables. We must exclude them.
    bounds_vars = set()
    for k in dataset.variables:
        var = dataset[k]

        # Bounds variable as recommended through CF conventions
        bounds_k = var.attrs.get('bounds')
        if bounds_k is not None and bounds_k in dataset:
            bounds_vars.add(bounds_k)

        # Bounds variable by naming convention,
        # e.g. "lon_bnds" or "lat_bounds"
        k_splits = str(k).rsplit("_", maxsplit=1)
        if len(k_splits) == 2:
            k_base, k_suffix = k_splits
            if k_suffix in ('bnds', 'bounds') and k_base in dataset:
                bounds_vars.add(k)

    potential_coord_vars = []

    # First consider any CF global attribute "coordinates"
    coordinates = dataset.attrs.get('coordinates')
    if coordinates is not None:
        for var_name in coordinates.split():
            if _is_potential_coord_var(dataset, bounds_vars, var_name):
                potential_coord_vars.append(var_name)

    # Then consider any other 1D/2D variables
    for var_name in dataset.variables:
        if var_name not in potential_coord_vars \
                and _is_potential_coord_var(dataset, bounds_vars, var_name):
            potential_coord_vars.append(var_name)

    return potential_coord_vars


def _is_potential_coord_var(dataset: xr.Dataset,
                            bounds_var_names: Set[str],
                            var_name: Hashable) -> bool:
    if var_name in dataset:
        var = dataset[var_name]
        return var.ndim in (1, 2) and var_name not in bounds_var_names
    return False


def _find_dataset_tile_size(dataset: xr.Dataset,
                            x_dim_name: Hashable,
                            y_dim_name: Hashable) \
        -> Optional[Tuple[int, int]]:
    """Find the most likely tile size in *dataset*"""
    dataset_chunks = get_dataset_chunks(dataset)
    tile_width = dataset_chunks.get(x_dim_name)
    tile_height = dataset_chunks.get(y_dim_name)
    if tile_width is not None and tile_height is not None:
        return tile_width, tile_height
    return None


def add_spatial_ref(dataset_store: zarr.convenience.StoreLike,
                    crs: pyproj.CRS,
                    crs_var_name: Optional[str] = 'spatial_ref',
                    xy_dim_names: Optional[Tuple[str, str]] = None):
    """
    Helper function that allows adding a spatial reference to an
    existing Zarr dataset.

    :param dataset_store: The dataset's existing Zarr store or path.
    :param crs: The spatial coordinate reference system.
    :param crs_var_name: The name of the variable that will hold the
        spatial reference. Defaults to "spatial_ref".
    :param xy_dim_names: The names of the x and y dimensions.
        Defaults to ("x", "y").
    """
    assert_instance(dataset_store, (MutableMapping, str), name='group_store')
    assert_instance(crs_var_name, str, name='crs_var_name')
    x_dim_name, y_dim_name = xy_dim_names or ('x', 'y')

    spatial_attrs = crs.to_cf()
    spatial_attrs['_ARRAY_DIMENSIONS'] = []  # Required by xarray
    group = zarr.open(dataset_store, mode='r+')
    spatial_ref = group.array(crs_var_name,
                              0,
                              shape=(),
                              dtype=np.uint8,
                              fill_value=0)
    spatial_ref.attrs.update(**spatial_attrs)

    for item_name, item in group.items():
        if item_name != crs_var_name:
            dims = item.attrs.get('_ARRAY_DIMENSIONS')
            if dims and len(dims) >= 2 \
                    and dims[-2] == y_dim_name \
                    and dims[-1] == x_dim_name:
                item.attrs['grid_mapping'] = crs_var_name

    zarr.convenience.consolidate_metadata(dataset_store)



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Dict, Set

import xarray as xr

ANCILLARY_VAR_NAME_PREFIXES_TO_STANDARD_NAME_MODIFIERS = [
    ('std', 'standard_error'),
    ('count', 'number_of_observations'),
]


def find_ancillary_var_names(dataset: xr.Dataset,
                             var_name: str,
                             same_shape: bool = False,
                             same_dims: bool = False) -> Dict[str, Set[str]]:
    variable = dataset.data_vars.get(var_name)

    results = {}

    # Check for CF compatibility according to
    # http://cfconventions.org/Data/cf-conventions/cf-conventions-1.7/cf-conventions.html#ancillary-data
    #
    if variable is not None and 'ancillary_variables' in variable.attrs:
        ancillary_var_names = variable.attrs['ancillary_variables'].split(" ")
        for ancillary_var_name in ancillary_var_names:
            if ancillary_var_name in dataset.data_vars:
                ancillary_var = dataset.data_vars[ancillary_var_name]
                if (not same_shape or variable.shape == ancillary_var.shape) \
                        and (not same_dims or variable.dims == variable.dims):
                    standard_name_modifier = _get_standard_name_modifier(variable, ancillary_var) or ''
                    if standard_name_modifier not in results:
                        results[standard_name_modifier] = set()
                    results[standard_name_modifier].add(ancillary_var_name)

    if variable is not None and not results:
        #
        # Check for less strict CF compatibility (missing attribute 'ancillary_variables')
        #
        if 'standard_name' in variable.attrs:
            for ancillary_var_name, ancillary_var in dataset.data_vars.items():
                if ancillary_var is variable:
                    continue
                standard_name_modifier = _get_standard_name_modifier(variable, ancillary_var)
                if standard_name_modifier is not None:
                    if standard_name_modifier not in results:
                        results[standard_name_modifier] = set()
                    results[standard_name_modifier].add(ancillary_var_name)

    if not results:
        #
        # Search for variables with xcube-specific prefixes that indicate uncertainty:
        #
        for prefix, standard_name_modifier in ANCILLARY_VAR_NAME_PREFIXES_TO_STANDARD_NAME_MODIFIERS:
            ancillary_var_name = f"{var_name}_{prefix}"
            if ancillary_var_name in dataset.data_vars:
                if standard_name_modifier not in results:
                    results[standard_name_modifier] = set()
                results[standard_name_modifier].add(ancillary_var_name)

    return results


def _get_standard_name_modifier(variable, ancillary_var):
    """
    See CF Conventions v 1.7, Appendix C: Standard Name Modifiers
    """
    if 'standard_name' in ancillary_var.attrs:
        ancillary_var_std_name = ancillary_var.attrs['standard_name']
        ancillary_var_std_name_parts = ancillary_var_std_name.split(" ")
        if len(ancillary_var_std_name_parts) == 2 \
                and ancillary_var_std_name_parts[0] == variable.attrs.get('standard_name'):
            return ancillary_var_std_name_parts[-1]
    return None



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import json
from functools import cached_property
from itertools import filterfalse
from string import Template
from typing import Optional, Mapping, Dict, Any, Set, Union

import jwt
import jwt.algorithms
import requests

from xcube.server.api import ApiContext
from xcube.server.api import ApiError
from .config import AuthConfig


class AuthContext(ApiContext):

    @cached_property
    def auth_config(self) -> Optional[AuthConfig]:
        return AuthConfig.from_config(self.config)

    @property
    def can_authenticate(self) -> bool:
        """Test whether the user can authenticate.
        Even if authentication service is configured, user authentication
        may still be optional. In this case the server will publish
        the resources configured to be free for everyone.
        """
        return self.auth_config is not None

    @property
    def must_authenticate(self) -> bool:
        """
        Test whether the user must authenticate.
        """
        return self.auth_config is not None and self.auth_config.is_required

    @cached_property
    def jwks(self) -> Dict[str, Any]:
        response = requests.get(self.auth_config.well_known_jwks)
        return json.loads(response.content)

    @cached_property
    def jwks(self):
        assert self.auth_config is not None
        jwks_uri = self.auth_config.well_known_jwks
        openid_config_uri = self.auth_config.well_known_oid_config
        response = requests.get(openid_config_uri)
        if response.ok:
            openid_config = json.loads(response.content)
            if openid_config and 'jwks_uri' in openid_config:
                jwks_uri = openid_config['jwks_uri']
        response = requests.get(jwks_uri)
        if response.ok:
            return json.loads(response.content)
        # TODO (forman): convert into ApiError
        response.raise_for_status()

    def get_granted_scopes(self, request_headers: Mapping[str, str]) \
            -> Optional[Set[str]]:
        must_authenticate = self.must_authenticate
        id_token = self.get_id_token(request_headers,
                                     require_auth=must_authenticate)
        permissions = None
        if id_token:
            permissions = id_token.get('permissions')
            if not isinstance(permissions, (list, tuple)):
                scope = id_token.get('scope')
                if isinstance(scope, str):
                    permissions = scope.split(' ')
            if permissions is not None:
                permissions = self._interpolate_permissions(id_token,
                                                            permissions)
        return permissions

    def get_id_token(self,
                     request_headers: Mapping[str, str],
                     require_auth: bool = False) \
            -> Optional[Mapping[str, str]]:
        """Decode the access token and verifies it."""

        access_token = self.get_access_token(request_headers,
                                             require_auth=require_auth)
        if access_token is None:
            return None

        auth_config = self.auth_config
        if auth_config is None:
            # Ignore access token
            return None

        # With auth_config and access_token, we expect authorization
        # to work. From here on we raise, if anything fails.

        # Get the unverified header of the access token
        try:
            unverified_header = jwt.get_unverified_header(access_token)
        except jwt.InvalidTokenError:
            unverified_header = None
        if not unverified_header \
                or not unverified_header.get("kid") \
                or not unverified_header.get("alg"):
            # "alg" should be "RS256" or "HS256" or others
            raise ApiError.BadRequest(
                "Invalid header."
                " A signed JWT Access Token is expected."
            )

        # The key identifier of the access token which we must validate.
        access_token_kid = unverified_header["kid"]

        # Get JSON Web Token (JWK) Keys
        jwks = self.jwks

        # Find access_token_kid in JWKS to obtain rsa_key
        rsa_key = None
        for key in jwks["keys"]:
            if key["kid"] == access_token_kid:
                rsa_key = {
                    "kty": key["kty"],
                    "kid": key["kid"],
                    "use": key["use"],
                    "n": key["n"],
                    "e": key["e"]
                }
                break
        if rsa_key is None:
            raise ApiError.BadRequest(
                "Invalid header. Unable to find appropriate key in JWKS."
            )

        # Now we are ready to decode the access token
        try:
            id_token = jwt.decode(
                access_token,
                jwt.algorithms.RSAAlgorithm.from_jwk(rsa_key),
                issuer=auth_config.authority,
                audience=auth_config.audience,
                algorithms=auth_config.algorithms
            )
        except jwt.PyJWTError as e:
            raise ApiError.BadRequest(
                f"Failed to decode access token: {e}"
            ) from e

        return id_token

    @classmethod
    def get_access_token(cls,
                         request_headers: Mapping[str, str],
                         require_auth: bool = False) -> Optional[str]:
        """Obtain the access token from the Authorization Header
        """
        # noinspection PyUnresolvedReferences
        auth = request_headers.get("Authorization", None)
        if not auth:
            if require_auth:
                raise ApiError.Unauthorized(
                    "Authorization header is expected."
                )
            return None

        parts = auth.split()

        if parts[0].lower() != "bearer":
            raise ApiError.BadRequest(
                'Invalid header.'
                ' Authorization header must start with "Bearer".'
            )
        elif len(parts) == 1:
            raise ApiError.BadRequest(
                "Invalid header."
                " Bearer token not found."
            )
        elif len(parts) > 2:
            raise ApiError.BadRequest(
                "Invalid header."
                " Authorization header must be Bearer token."
            )

        return parts[1]

    def _interpolate_permissions(self,
                                 id_token: Mapping[str, Any],
                                 permissions: Union[list, tuple]):
        predicate = self._is_template_permission

        plain_permissions = set(filterfalse(predicate, permissions))
        if len(plain_permissions) == len(permissions):
            return plain_permissions

        templ_permissions = filter(predicate, permissions)
        id_mapping = self._get_template_dict(id_token)
        return plain_permissions.union(
            set(Template(permission).safe_substitute(id_mapping)
                for permission in templ_permissions)
        )

    @staticmethod
    def _is_template_permission(permission: str) -> bool:
        return '$' in permission

    @staticmethod
    def _get_template_dict(id_token: Mapping[str, Any]) -> Dict[str, str]:
        d = {k: v
             for k, v in id_token.items()
             if isinstance(v, str)}
        if 'username' not in d and 'preferred_username' in d:
            d['username'] = d['preferred_username']
        return d



# The MIT License (MIT)
# Copyright (c) 2023 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
import os
import re
import tempfile
from typing import Mapping, Sequence, Optional, Any, Literal, NamedTuple, Union

import numpy as np
import pyproj
import rasterio
import xarray as xr

from xcube.core.gridmapping import GridMapping
from xcube.core.resampling import resample_in_space
from xcube.server.api import ApiError
from xcube.util.timeindex import ensure_time_index_compatible
from xcube.webapi.datasets.context import DatasetsContext
from xcube.webapi.ows.coverages.request import CoverageRequest
from xcube.webapi.ows.coverages.scaling import CoverageScaling
from xcube.webapi.ows.coverages.util import get_h_dim, get_v_dim


def get_coverage_as_json(ctx: DatasetsContext, collection_id: str):
    """
    Return a JSON representation of the specified coverage

    Currently, the range set component is omitted.

    :param ctx: a dataset context
    :param collection_id: the ID of the collection providing the coverage
    :return: a JSON representation of the coverage
    """
    return {
        'id': collection_id,
        'type': 'CoverageByDomainAndRange',
        'envelope': get_collection_envelope(ctx, collection_id),
        'domainSet': get_coverage_domainset(ctx, collection_id),
        'rangeSet': {
            'type': 'RangeSet',
            # TODO: Wait for next update to API specification before
            #  implementing the data block -- not clear yet whether this
            #  is being deprecated along with the rangeSet endpoint.
            'dataBlock': {'type': 'VDataBlock', 'values': ['TODO']},
        },
        'rangeType': get_coverage_rangetype(ctx, collection_id),
        'metadata': get_collection_metadata(ctx, collection_id),
    }


def get_coverage_data(
    ctx: DatasetsContext,
    collection_id: str,
    query: Mapping[str, Sequence[str]],
    content_type: str,
) -> tuple[Optional[bytes], list[float], pyproj.CRS]:
    """
    Return coverage data from a dataset

    This method currently returns coverage data from a dataset as either
    TIFF or NetCDF. The bbox, datetime, and properties parameters are
    currently handled.

    :param ctx: a datasets context
    :param collection_id: the dataset from which to return the coverage
    :param query: the HTTP query parameters
    :param content_type: the MIME type of the desired output format
    :return: A tuple consisting of: (1) the coverage as bytes in the requested
             output format, or None if the requested output format is not
             supported; (2) the bounding box of the returned data, respecting
             the axis ordering of the CRS (e.g. latitude first for EPSG:4326);
             (3) the CRS of the returned dataset and bounding box
    """

    ds = get_dataset(ctx, collection_id)

    try:
        request = CoverageRequest(query)
    except ValueError as e:
        raise ApiError.BadRequest(str(e))

    # See https://docs.ogc.org/DRAFTS/19-087.html
    native_crs = get_crs_from_dataset(ds)
    final_crs = request.crs if request.crs is not None else native_crs
    bbox_crs = request.bbox_crs
    subset_crs = request.subset_crs

    if request.properties is not None:
        ds = _apply_properties(collection_id, ds, request.properties)

    # https://docs.ogc.org/DRAFTS/19-087.html#datetime-parameter-subset-requirements
    # requirement 7D: "If a datetime parameter is specified requesting a
    # coverage without any temporal dimension, the parameter SHALL either be
    # ignored, or a 4xx client error generated." We choose to ignore it.
    if request.datetime is not None and 'time' in ds.variables:
        if isinstance(request.datetime, tuple):
            time_slice = slice(*request.datetime)
            time_slice = ensure_time_index_compatible(ds, time_slice)
            ds = ds.sel(time=time_slice)
        else:
            timespec = ensure_time_index_compatible(ds, request.datetime)
            ds = ds.sel(time=timespec, method='nearest').squeeze()

    if request.subset is not None:
        subset_bbox, ds = _apply_subsetting(ds, request.subset, subset_crs)
    else:
        subset_bbox = None

    if request.bbox is not None:
        ds = _apply_bbox(ds, request.bbox, bbox_crs, always_xy=False)

    # Do a provisional size check with an approximate scaling before attempting
    # to determine a grid mapping, so the client gets a comprehensible error
    # if the coverage is empty or too large.
    _assert_coverage_size_ok(CoverageScaling(request, final_crs, ds))

    transformed_gm = source_gm = GridMapping.from_dataset(ds, crs=native_crs)
    if native_crs != final_crs:
        transformed_gm = transformed_gm.transform(final_crs).to_regular()
    if transformed_gm is not source_gm:
        # We can't combine the scaling operation with this CRS transformation,
        # since the size may end up wrong after the re-application of bounding
        # boxes below.
        ds = resample_in_space(
            ds, source_gm=source_gm, target_gm=transformed_gm
        )

    if native_crs != final_crs:
        # If we've resampled into a new CRS, the transformed native-CRS
        # bbox[es] may have been too big, so we re-crop in the final CRS.
        if request.bbox is not None:
            ds = _apply_bbox(ds, request.bbox, bbox_crs, always_xy=False)
        if subset_bbox is not None:
            ds = _apply_bbox(ds, subset_bbox, subset_crs, always_xy=True)

    # Apply final size check and scaling operation after bbox, subsetting,
    # and CRS transformation, to make sure that the final size is correct.
    scaling = CoverageScaling(request, final_crs, ds)
    _assert_coverage_size_ok(scaling)
    if scaling.factor != (1, 1):
        cropped_gm = GridMapping.from_dataset(ds, crs=final_crs)
        scaled_gm = scaling.apply(cropped_gm)
        ds = resample_in_space(ds, source_gm=cropped_gm, target_gm=scaled_gm)

    ds.rio.write_crs(final_crs, inplace=True)
    for var in ds.data_vars.values():
        var.attrs.pop('grid_mapping', None)

    # check: rename axes to match final CRS?

    media_types = dict(
        tiff={'geotiff', 'image/tiff', 'application/x-geotiff'},
        png={'png', 'image/png'},
        netcdf={'netcdf', 'application/netcdf', 'application/x-netcdf'},
    )
    if content_type in media_types['tiff']:
        content = dataset_to_image(ds, 'tiff', final_crs)
    elif content_type in media_types['png']:
        content = dataset_to_image(ds, 'png', final_crs)
    elif content_type in media_types['netcdf']:
        content = dataset_to_netcdf(ds)
    else:
        # It's expected that the caller (server API handler) will catch
        # unhandled types, but we may as well do the right thing if any
        # do slip through.
        raise ApiError.UnsupportedMediaType(
            f'Unsupported media type {content_type}. '
            + 'Available media types: '
            + ', '.join(
                [type_ for value in media_types.values() for type_ in value]
            )
        )
    final_bbox = get_bbox_from_ds(ds)
    if not is_xy_order(final_crs):
        final_bbox = final_bbox[1], final_bbox[0], final_bbox[3], final_bbox[2]
    return content, final_bbox, final_crs


def _apply_properties(collection_id, ds, properties):
    requested_vars = set(properties)
    data_vars = set(map(
        # Filter out 0-dimensional variables (usually grid mapping variables)
        str, {k: v for k, v in ds.data_vars.items() if v.dims != ()}
    ))
    unrecognized_vars = requested_vars - data_vars
    if unrecognized_vars == set():
        ds = ds.drop_vars(
            list(data_vars - requested_vars - {'crs', 'spatial_ref'})
        )
    else:
        raise ApiError.BadRequest(
            f'The following properties are not present in the coverage '
            f'{collection_id}: {", ".join(unrecognized_vars)}'
        )
    return ds


def _assert_coverage_size_ok(scaling: CoverageScaling):
    size_limit = 4000 * 4000  # TODO make this configurable
    x, y = scaling.size
    if (x * y) > size_limit:
        raise ApiError.ContentTooLarge(
            f'Requested coverage is too large:' f'{x} × {y} > {size_limit}.'
        )


_IndexerTuple = NamedTuple(
    'Indexers',
    [
        ('indices', dict[str, Any]),  # non-geographic single-valued specifiers
        ('slices', dict[str, slice]),  # non-geographic range specifiers
        ('x', Optional[Union[float, tuple[float, float]]]),  # x or longitude
        ('y', Optional[Union[float, tuple[float, float]]]),  # y or latitude
    ],
)


def _apply_subsetting(
    ds: xr.Dataset, subset_spec: dict, subset_crs: pyproj.CRS
) -> tuple[list[float], xr.Dataset]:
    indexers = _parse_subset_specifier(subset_spec, ds)

    # TODO: for geographic subsetting, also handle single-value (non-slice)
    #  indices and half-open slices.
    bbox = None
    if (indexers.x, indexers.y) != (None, None):
        bbox, ds = _apply_geographic_subsetting(ds, subset_crs, indexers)
    if indexers.slices:
        ds = ds.sel(indexers=indexers.slices)
    if indexers.indices:
        ds = ds.sel(indexers=indexers.indices, method='nearest')

    return bbox, ds


def _parse_subset_specifier(
    subset_spec: dict[str, Union[str, tuple[Optional[str], Optional[str]]]],
    ds: xr.Dataset,
) -> _IndexerTuple:
    specifiers = {}
    for axis, value in subset_spec.items():
        if isinstance(value, str):  # single value
            if axis == 'time':
                specifiers[axis] = ensure_time_index_compatible(ds, value)
            else:
                try:
                    # Parse to float if possible
                    specifiers[axis] = float(value)
                except ValueError:
                    specifiers[axis] = value
        else:  # range
            low, high = value
            low = None if low == '*' else low
            high = None if high == '*' else high
            if axis == 'time':
                specifiers[axis] = ensure_time_index_compatible(
                    ds, slice(low, high)
                )
            else:
                low = float(low)
                high = float(high)
                if axis.lower()[:3] in ['y', 'n', 'nor', 'lat'] and high < low:
                    low, high = high, low
                specifiers[axis] = low, high

    # Find and extract geographic parameters, if any. These have to be
    # handled specially, since they refer to axis names in the subsetting
    # CRS rather than dimension names in the dataset. For now, we actually
    # don't check the CRS axis names, but just look for parameters with
    # appropriate names (x, y, lat, long, etc.).
    x_param, y_param = _find_geographic_parameters(list(specifiers))
    x_value = specifiers.pop(x_param) if x_param is not None else None
    y_value = specifiers.pop(y_param) if y_param is not None else None

    # Separate index and slice (i.e. single-value and range) specifiers
    indices = {k: v for k, v in specifiers.items() if not isinstance(v, slice)}
    slices = {k: v for k, v in specifiers.items() if isinstance(v, slice)}

    return _IndexerTuple(indices, slices, x_value, y_value)


def _apply_geographic_subsetting(
    ds, subset_crs, indexers
) -> tuple[list[float], xr.Dataset]:
    # NB: We use xy axis ordering for the bounding boxes throughout this
    # function, regardless of what's specified in the CRSs.

    # 1. transform native extent to a whole-dataset bbox in subset_crs.
    # We'll use this to fill in "full extent" values if geographic
    # subsetting is only specified in one dimension.
    full_bbox_native = get_bbox_from_ds(ds)
    native_crs = get_crs_from_dataset(ds)
    full_bbox_subset_crs = transform_bbox(
        full_bbox_native, native_crs, subset_crs
    )

    # 2. Find horizontal and/or vertical ranges in indexers, falling back to
    # values from whole-dataset bbox if a complete bbox is not specified.
    x0, x1 = (
        indexers.x
        if indexers.x is not None
        else (full_bbox_subset_crs[0], full_bbox_subset_crs[2])
    )
    y0, y1 = (
        indexers.y
        if indexers.y is not None
        else (full_bbox_subset_crs[1], full_bbox_subset_crs[3])
    )

    # 3. Using the ranges determined from the indexers and whole-dataset bbox,
    # construct the requested bbox in the subsetting CRS.
    bbox_subset_crs = [x0, y0, x1, y1]

    # 4. Transform requested bbox from subsetting CRS to dataset-native CRS.
    bbox_native_crs = transform_bbox(bbox_subset_crs, subset_crs, native_crs)

    # 6. Apply the dataset-native bbox using sel, making sure that y/latitude
    # slice has the same ordering as the corresponding co-ordinate.
    h_dim = get_h_dim(ds)
    v_dim = get_v_dim(ds)
    ds = ds.sel(
        indexers={
            h_dim: slice(bbox_native_crs[0], bbox_native_crs[2]),
            v_dim: slice(
                *_correct_inverted_y_range_if_necessary(
                    ds, v_dim, (bbox_native_crs[1], bbox_native_crs[3])
                )
            ),
        }
    )
    return bbox_subset_crs, ds


def get_bbox_from_ds(ds: xr.Dataset):
    h, v = ds[get_h_dim(ds)], ds[get_v_dim(ds)]
    bbox = list(map(float, [h[0], v[0], h[-1], v[-1]]))
    _ensure_bbox_y_ascending(bbox)
    return bbox


def _find_geographic_parameters(
    names: list[str],
) -> tuple[Optional[str], Optional[str]]:
    x, y = None, None
    for name in names:
        if name.lower()[:3] in ['x', 'e', 'eas', 'lon']:
            x = name
        if name.lower()[:3] in ['y', 'n', 'nor', 'lat']:
            y = name
    return x, y


def transform_bbox(
    bbox: list[float], source_crs: pyproj.CRS, dest_crs: pyproj.CRS
) -> list[float]:
    if source_crs == dest_crs:
        return bbox
    transformer = pyproj.Transformer.from_crs(
        source_crs, dest_crs, always_xy=True
    )
    bbox_ = bbox.copy()
    _ensure_bbox_y_ascending(bbox_)
    return list(transformer.transform_bounds(*bbox_))


def _apply_bbox(
    ds: xr.Dataset, bbox: list[float], bbox_crs: pyproj.CRS, always_xy: bool
):
    native_crs = get_crs_from_dataset(ds)

    if native_crs != bbox_crs:
        transformer = pyproj.Transformer.from_crs(
            bbox_crs, native_crs, always_xy=always_xy
        )
        _ensure_bbox_y_ascending(bbox, always_xy or is_xy_order(bbox_crs))
        bbox = transformer.transform_bounds(*bbox)
    h_dim = get_h_dim(ds)
    v_dim = get_v_dim(ds)
    x0, y0, x1, y1 = (
        (0, 1, 2, 3)
        if (always_xy or is_xy_order(native_crs))
        else (1, 0, 3, 2)
    )
    v_slice = _correct_inverted_y_range_if_necessary(
        ds, v_dim, (bbox[y0], bbox[y1])
    )
    ds = ds.sel({h_dim: slice(bbox[x0], bbox[x1]), v_dim: slice(*v_slice)})
    return ds


def _ensure_bbox_y_ascending(bbox: list, xy_order: bool = True):
    y0, y1 = (1, 3) if xy_order else (0, 2)
    if bbox[y0] > bbox[y1]:
        bbox[y0], bbox[y1] = bbox[y1], bbox[y0]


def _correct_inverted_y_range_if_necessary(
    ds: xr.Dataset, axis: str, range_: tuple[float, float]
) -> tuple[float, float]:
    x0, x1 = range_
    # Make sure latitude slice direction matches axis direction.
    # (For longitude, a descending-order slice is valid.)
    if (
        None not in range_
        and axis[:3].lower() in {'lat', 'nor', 'y'}
        and (x0 < x1) != (ds[axis][0] < ds[axis][-1])
    ):
        x0, x1 = x1, x0
    return x0, x1


def dataset_to_image(
    ds: xr.Dataset,
    image_format: Literal['png', 'tiff'] = 'png',
    crs: pyproj.CRS = None,
) -> bytes:
    """
    Return an in-memory bitmap (TIFF or PNG) representing a dataset

    :param ds: a dataset
    :param image_format: image format to generate ("png" or "tiff")
    :param crs: CRS of the dataset
    :return: TIFF-formatted bytes representing the dataset
    """

    if image_format == 'png':
        for var in ds.data_vars:
            # rasterio's PNG driver only supports these data types.
            if ds[var].dtype not in {np.uint8, np.uint16}:
                ds[var] = ds[var].astype(np.uint16, casting='unsafe')

    ds = ds.squeeze()

    with tempfile.TemporaryDirectory() as tempdir:
        path = os.path.join(tempdir, 'out.' + image_format)
        # Make dataset representable in an image format by discarding
        # additional variables and dimensions.
        ds = ds.drop_vars(
            names=['crs', 'spatial_ref'], errors='ignore'
        ).squeeze()
        if len(ds.data_vars) == 1:
            ds[list(ds.data_vars)[0]].rio.to_raster(path)
        else:
            ds.rio.to_raster(path)
        if crs is not None:
            with rasterio.open(path, mode='r+') as src:
                src.crs = crs
        with open(path, 'rb') as fh:
            data = fh.read()
    return data


def dataset_to_netcdf(ds: xr.Dataset) -> bytes:
    """
    Return an in-memory NetCDF representing a dataset

    :param ds: a dataset
    :return: NetCDF-formatted bytes representing the dataset
    """
    with tempfile.TemporaryDirectory() as tempdir:
        path = os.path.join(tempdir, 'out.nc')
        ds.to_netcdf(path)
        with open(path, 'rb') as fh:
            data = fh.read()
    return data


def get_coverage_domainset(ctx: DatasetsContext, collection_id: str):
    """
    Return the domain set of a dataset-backed coverage

    The domain set is the set of input parameters (e.g. geographical extent,
    time span) for which a coverage is defined.

    :param ctx: a datasets context
    :param collection_id: the dataset for which to return the domain set
    :return: a dictionary representing an OGC API - Coverages domain set
    """
    ds = get_dataset(ctx, collection_id)
    grid_limits = dict(
        type='GridLimits',
        srsName=f'http://www.opengis.net/def/crs/OGC/0/Index{len(ds.dims)}D',
        axisLabels=list(ds.dims),
        axis=[_get_grid_limits_axis(ds, dim) for dim in ds.dims],
    )
    grid = dict(
        type='GeneralGridCoverage',
        srsName=get_crs_from_dataset(ds).to_string(),
        axisLabels=list(ds.dims.keys()),
        axis=_get_axes_properties(ds),
        gridLimits=grid_limits,
    )
    return dict(type='DomainSet', generalGrid=grid)


def get_collection_metadata(ctx: DatasetsContext, collection_id: str):
    """
    Return a metadata dictionary for a dataset

    The metadata is taken directly from the dataset attributes.

    :param ctx: a datasets context
    :param collection_id: the dataset for which to return the metadata
    :return: a dictionary of metadata keys and values
    """
    ds = get_dataset(ctx, collection_id)
    return ds.attrs


def get_dataset(ctx: DatasetsContext, collection_id: str):
    """
    Get a dataset from a datasets context

    :param ctx: a datasets context
    :param collection_id: the ID of a dataset in the context
    :return: the dataset
    """
    ml_dataset = ctx.get_ml_dataset(collection_id)
    ds = ml_dataset.get_dataset(0)
    assert isinstance(ds, xr.Dataset)
    return ds


def _get_axes_properties(ds: xr.Dataset) -> list[dict]:
    return [_get_axis_properties(ds, dim) for dim in ds.dims]


def _get_axis_properties(ds: xr.Dataset, dim: str) -> dict[str, Any]:
    axis = ds.coords[dim]
    if np.issubdtype(axis.dtype, np.datetime64):
        lower_bound = np.datetime_as_string(axis[0])
        upper_bound = np.datetime_as_string(axis[-1])
    else:
        lower_bound, upper_bound = axis[0].item(), axis[-1].item()
    return dict(
        type='RegularAxis',
        axisLabel=dim,
        lowerBound=lower_bound,
        upperBound=upper_bound,
        resolution=abs((axis[-1] - axis[0]).item() / len(axis)),
        uomLabel=get_units(ds, dim),
    )


def _get_grid_limits_axis(ds: xr.Dataset, dim: str) -> dict[str, Any]:
    return dict(
        type='IndexAxis', axisLabel=dim, lowerBound=0, upperBound=len(ds[dim])
    )


def get_units(ds: xr.Dataset, dim: str) -> str:
    coord = ds.coords[dim]
    if hasattr(coord, 'attrs') and 'units' in coord.attrs:
        return coord.attrs['units']
    if np.issubdtype(coord, np.datetime64):
        return np.datetime_data(coord)[0]
    # TODO: as a fallback for spatial axes, we could try matching dimensions
    #  to CRS axes and take the unit from the CRS definition.
    return 'unknown'


def get_crs_from_dataset(ds: xr.Dataset) -> pyproj.CRS:
    """
    Return the CRS of a dataset as a string. The CRS is taken from the
    metadata of the crs or spatial_ref variables, if available.
    "EPSG:4326" is used as a fallback.

    :param ds: a dataset
    :return: a string representation of the dataset's CRS, or "EPSG:4326"
             if the CRS cannot be determined
    """
    for var_name in 'crs', 'spatial_ref':
        if var_name in ds.variables:
            var = ds[var_name]
            for attr_name in 'spatial_ref', 'crs_wkt':
                if attr_name in var.attrs:
                    crs_string = ds[var_name].attrs[attr_name]
                    return pyproj.CRS(crs_string)
    return pyproj.CRS('EPSG:4326')


def get_coverage_rangetype(
    ctx: DatasetsContext, collection_id: str
) -> dict[str, list]:
    """Return the range type of a dataset

    The range type describes the data types of the dataset's variables
    using a format defined in https://docs.ogc.org/is/09-146r6/09-146r6.html

    :param ctx: datasets context
    :param collection_id: ID of the dataset in the supplied context
    :return: a dictionary representing the specified dataset's range type
    """
    return get_coverage_rangetype_for_dataset(get_dataset(ctx, collection_id))


def get_coverage_rangetype_for_dataset(ds) -> dict[str, list]:
    """Return the range type of a dataset

    The range type describes the data types of the dataset's variables
    using a format defined in https://docs.ogc.org/is/09-146r6/09-146r6.html

    :param ds: a dataset
    :return: a dictionary representing the supplied dataset's range type
    """
    result = dict(type='DataRecord', field=[])
    for var_name, variable in ds.data_vars.items():
        if variable.dims == ():
            # A 0-dimensional variable is probably a grid mapping variable;
            # in any case, it doesn't have the dimensions of the cube, so
            # isn't part of the range.
            continue
        result['field'].append(
            dict(
                type='Quantity',
                name=var_name,
                description=get_dataarray_description(variable),
                encodingInfo=dict(
                    dataType=dtype_to_opengis_datatype(variable.dtype)
                ),
            )
        )
    return result


def dtype_to_opengis_datatype(dt: np.dtype) -> str:
    """
    Convert a NumPy dtype to an equivalent OpenGIS type identifier string.

    :param dt: a NumPy dtype
    :return: an equivalent OpenGIS type identifier string, or an empty string
             if the dtype is not recognized
    """
    nbits = 8 * np.dtype(dt).itemsize
    int_size_map = {8: 'Byte', 16: 'Short', 32: 'Int', 64: 'Long'}
    prefix = 'http://www.opengis.net/def/dataType/OGC/0/'
    if np.issubdtype(dt, np.floating):
        opengis_type = f'{prefix}float{nbits}'
    elif np.issubdtype(dt, np.signedinteger):
        opengis_type = f'{prefix}signed{int_size_map[nbits]}'
    elif np.issubdtype(dt, np.unsignedinteger):
        opengis_type = f'{prefix}unsigned{int_size_map[nbits]}'
    elif 'datetime64' in str(dt):
        opengis_type = 'http://www.opengis.net/def/bipm/UTC'
    else:
        opengis_type = ''  # TODO decide what to do in this case
    return opengis_type


def get_dataarray_description(da: xr.DataArray) -> str:
    """
    Return a string describing a DataArray, either from an attribute or,
    as a fallback, from its name attribute.

    :param da: a DataArray
    :return: a string describing the DataArray
    """
    if hasattr(da, 'attrs'):
        for attr in ['description', 'long_name', 'standard_name', 'name']:
            if attr in da.attrs:
                return da.attrs[attr]
    return str(da.name)


def get_collection_envelope(ds_ctx, collection_id):
    """
    Return the OGC API - Coverages envelope of a dataset.

    The envelope comprises the extents of all the dataset's dimensions.

    :param ds_ctx: a datasets context
    :param collection_id: a dataset ID within the given context
    :return: the envelope of the specified dataset
    """
    ds = get_dataset(ds_ctx, collection_id)
    return {
        'type': 'EnvelopeByAxis',
        'srsName': get_crs_from_dataset(ds).to_string(),
        'axisLabels': list(ds.dims.keys()),
        'axis': _get_axes_properties(ds),
    }


def is_xy_order(crs: pyproj.CRS) -> bool:
    """Try to determine whether a CRS has x-y axis order"""
    x_index = None
    y_index = None
    x_re = re.compile('^x|lon|east', flags=re.IGNORECASE)
    y_re = re.compile('^y|lat|north', flags=re.IGNORECASE)

    for i, axis in enumerate(crs.axis_info):
        for prop in 'name', 'abbrev', 'direction':
            if x_re.search(getattr(axis, prop)):
                x_index = i
            elif y_re.search(getattr(axis, prop)):
                y_index = i

    if x_index is not None and y_index is not None:
        return x_index < y_index
    else:
        return True  # assume xy



# The MIT License (MIT)
# Copyright (c) 2023 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import functools
import importlib
import inspect
import concurrent.futures
import datetime
from typing import Dict, Any, Optional, Callable
import traceback

import xarray as xr

from xcube.core.mldataset import MultiLevelDataset
from xcube.server.api import Context, ApiError
from xcube.webapi.common.context import ResourcesContext
from xcube.webapi.datasets.context import DatasetsContext
from xcube.webapi.places import PlacesContext
from xcube.constants import LOG
from .op.info import OpInfo
from .op.registry import OpRegistry
from .op.registry import OP_REGISTRY

# Register default operations:
importlib.import_module("xcube.webapi.compute.operations")

LocalExecutor = concurrent.futures.ThreadPoolExecutor

# TODO: we should create a module 'job' and define better job classes.
#   Here we use dicts for time being.

Job = Dict[str, Any]
Jobs = Dict[int, Job]

JobRequest = Dict[str, Any]

JobFuture = concurrent.futures.Future
JobFutures = Dict[int, JobFuture]

JOB_INIT = "init"
JOB_SCHEDULED = "scheduled"
JOB_STARTED = "started"
JOB_COMPLETED = "completed"
JOB_FAILED = "failed"
JOB_CANCELLED = "cancelled"

JOB_STATUSES = {
    JOB_INIT,
    JOB_SCHEDULED,
    JOB_STARTED,
    JOB_COMPLETED,
    JOB_FAILED,
    JOB_CANCELLED
}


class ComputeContext(ResourcesContext):
    """An xcube API server context for xcube compute operations

    Depends on the availability of `datasets` and `places` contexts
    in the same server context.
    """

    def __init__(self,
                 server_ctx: Context,
                 op_registry: OpRegistry = OP_REGISTRY):
        """Create a new compute context

        :param server_ctx: the current server context object
        :param op_registry: the registry of compute operations to use for
                            this context
        """
        super().__init__(server_ctx)
        self._datasets_ctx: DatasetsContext \
            = server_ctx.get_api_ctx("datasets")
        assert isinstance(self._datasets_ctx, DatasetsContext)
        self._places_ctx: PlacesContext \
            = server_ctx.get_api_ctx("places")
        assert isinstance(self._places_ctx, PlacesContext)

        self._op_registry = op_registry
        assert isinstance(self._op_registry, OpRegistry)

        compute_config = server_ctx.config.get("Compute", {})
        max_workers = compute_config.get("MaxWorkers", 3)

        self.next_job_id = 0
        self.jobs: Jobs = {}
        self.job_futures: JobFutures = {}
        self.job_executor = LocalExecutor(max_workers=max_workers,
                                          thread_name_prefix='xcube-job-')

    def on_dispose(self):
        self.job_executor.shutdown(cancel_futures=True)

    @property
    def op_registry(self) -> OpRegistry:
        """:return: the operation registry used by this compute context"""
        return self._op_registry

    @property
    def datasets_ctx(self) -> DatasetsContext:
        """:return: the datasets context used by this compute context"""
        return self._datasets_ctx

    @property
    def places_ctx(self) -> PlacesContext:
        """:return: the places context used by this compute context"""
        return self._places_ctx

    def schedule_job(self, job_request: JobRequest) -> Job:
        """Schedule a new job given by *job_request*,
        which is expected to be validated already.

        Job status transitions:

        init --> pending --> cancelled
                         --> failed
                         --> running --> completed
                                     --> failed
                                     --> cancelled
        """

        with self.rlock:
            job_id = self.next_job_id
            self.next_job_id += 1

        # Note, the order of following statements is crucial:

        # Create new job
        job = new_job(job_id, job_request)
        # Register new job
        self.jobs[job_id] = job
        set_job_status(job, JOB_SCHEDULED)
        # Schedule job
        job_future: JobFuture = \
            self.job_executor.submit(self._invoke_job, job_id)
        # Register new job future
        self.job_futures[job_id] = job_future
        # Notify when job is completed, failed, or cancelled.
        job_future.add_done_callback(
            functools.partial(self._handle_job_done, job_id)
        )

        return job

    def _handle_job_done(self, job_id: int, job_future: JobFuture):
        job = self.jobs.get(job_id)
        if job is None:
            return

        if job_future.cancelled():
            set_job_status(job, JOB_CANCELLED)
        else:
            error = job_future.exception()
            if error:
                set_job_status(job, JOB_FAILED, error=error)
            else:
                set_job_status(job, JOB_COMPLETED)

        # Make sure we get rid of reference
        self.job_futures.pop(job_id, None)

    def _invoke_job(self, job_id: int):
        """After successfully scheduling a job this method call is
        represented by a Future in self.job_futures.

        Since it is executed concurrently, it cannot return
        anything to the original requester.
        For the same reason, raising an ApiError will not set the
        HTTP response status.
        """
        job = self.jobs.get(job_id)
        if job is None:
            return

        job_request = job["request"]
        op_id = job_request["operationId"]
        parameters = job_request.get("parameters", {})
        output = job_request.get("output", {})

        set_job_status(job, JOB_STARTED)

        op = self.op_registry.get_op(op_id)
        parameters = self.get_effective_parameters(op, parameters)

        # Execute the operation!
        output_ds = op(**parameters)

        ds_id = self.datasets_ctx.add_dataset(
            output_ds,
            ds_id=output.get("datasetId"),
            title=output.get("title"),
            style=output.get("style"),
            color_mappings=output.get("colorMappings"),
        )

        set_job_result(job, {"datasetId": ds_id})

    def cancel_job(self, job_id: int) -> Job:
        """Cancel a scheduled job.

        :param job_id: the ID number of the job to be cancelled
        :return: details of the cancelled job as a string-keyed dictionary
        :raises ApiError: if the specified job cannot be found
        """
        job = self.jobs.get(job_id)
        if job is None:
            raise ApiError.NotFound(f'Job #{job_id} not found.')

        future = self.job_futures.pop(job_id, None)
        if future is not None and not future.done():
            set_job_status(job, JOB_CANCELLED)
            future.cancel()

        return job

    def get_effective_parameters(self,
                                 op: Callable,
                                 parameters: Dict[str, Any]):
        """Replace dataset names with datasets in operation parameters.

        This method takes a parameter dictionary for the invocation or an
        operation and returns a copy of the dictionary with any dataset
        names replaced by the actual referenced dataset. In other words,
        for any parameter where the operation expects a `Dataset` or
        `MultiLevelDataset` and the dictionary supplies a string, the string
        is replaced by the dataset with the corresponding name in this
        compute context’s dataset context.

        :param op: an operation
        :param parameters: parameters with which to execute the operation
        :return: a copy of the parameters, with dataset names replaced by
                 datasets
        """
        op_info = OpInfo.get_op_info(op)
        param_py_types = op_info.effective_param_py_types
        parameters = parameters.copy()
        for param_name, param_py_type in param_py_types.items():
            param_value = parameters.get(param_name)
            if isinstance(param_value, str) \
                    and inspect.isclass(param_py_type):
                if issubclass(param_py_type, xr.Dataset):
                    parameters[param_name] = \
                        self._datasets_ctx.get_dataset(param_value)
                elif issubclass(param_py_type, MultiLevelDataset):
                    parameters[param_name] = \
                        self._datasets_ctx.get_ml_dataset(param_value)
        return parameters


def new_job(job_id: int, job_request: JobRequest) -> Job:
    return {
        "jobId": job_id,
        "request": job_request,
        "state": {
            "status": "init"
        },
    }


def is_job_status(job: Job, status: str) -> bool:
    """Report whether a specified job has the specified status.

    :param job: a job specification (string-keyed dictionary)
    :param status: a string representing a recognized status
    :return: True if the status is recognized and the specified job has the
             specified status
    :raises ValueError: if the status is not recognized
    """
    _assert_valid_job_status(status)
    return job["state"]["status"] == status


def set_job_status(job: Job,
                   status: str,
                   error: Optional[BaseException] = None):
    """Set the status of a job.

    :param job: a job specification (string-keyed dictionary)
    :param status: a string representing a recognized status
    :param error: if supplied, annotate the job specification with information
                  from this exception
    :raises ValueError: if the status is not recognized
    """
    _assert_valid_job_status(status)
    job_id = job["jobId"]
    LOG.info(f"Job #{job_id} {status}.", exc_info=error)
    # Note, we could/should warn/raise on invalid state transitions
    job["state"]["status"] = status
    job["state"][f"{status}Time"] = datetime.datetime.utcnow().isoformat()
    if error is not None:
        job["state"]["error"] = {
            "message": str(error),
            "type": type(error).__name__,
            "traceback": traceback.extract_tb(error.__traceback__).format()
        }


def set_job_result(job: Job, result: Dict[str, Any]):
    """Set the result of a compute job.

    :param job: job specifier (string-keyed dictionary)
    :param result: results of the job
    """
    job["result"] = result


def _assert_valid_job_status(status):
    if status not in JOB_STATUSES:
        raise ValueError(f"illegal job status {status!r}")



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import click

__author__ = "Norman Fomferra (Brockmann Consult GmbH)"


@click.command(name='benchmark', hidden=True)
@click.argument('config')
@click.option('--repeats', '-R', metavar='REPEATS', type=int, default=1,
              help="Number of repetitions")
def benchmark(config: str, repeats: int):
    """
    Measure runtime performance of an executable.
    Record the processing time of the executable with varying parameters defined CONFIG.
    Repeat the executions REPEATS times.

    CONFIG is a YAML file that specifies the command to be executed and the parameter combinations
    to be benchmarked. The CONFIG

    \b
    command: my_exec --p1 ${param_1} --p2 ${param_2}
    params:
    - param_1
    - param_2
    param_1: [1, 2, 3]
    param_2: ['a', 'b']

    will benchmark all combinations of my_exec:

    \b
    my_exec --p1 1 --p2 a
    my_exec --p1 1 --p2 b
    my_exec --p1 2 --p2 a
    my_exec --p1 2 --p2 b
    my_exec --p1 3 --p2 a
    my_exec --p1 3 --p2 b
    """

    config_path = config
    repetition_count = repeats

    import yaml
    with open(config_path) as stream:
        config = yaml.safe_load(stream)

    import itertools

    command_template = config["command"]
    param_names = config["params"]

    param_values = [config[param_name] for param_name in param_names]

    import numpy as np
    param_values_combinations = list(itertools.product(*param_values))
    param_combination_count = len(param_values_combinations)
    times = np.ndarray((repetition_count, param_combination_count), dtype=np.float64)

    import subprocess
    import time
    import sys

    for repetition_index in range(repetition_count):
        for param_combination_index in range(param_combination_count):
            param_values = param_values_combinations[param_combination_index]
            params = list(zip(param_names, param_values))
            if isinstance(command_template, str):
                command = _apply_template(command_template, params)
            else:
                command = [_apply_template(arg_template, params) for arg_template in command_template]

            try:
                t0 = time.perf_counter()
                subprocess.check_call(command)
                t1 = time.perf_counter()
                times[repetition_index, param_combination_index] = t1 - t0
            except Exception as e:
                print(f"error: {e}", file=sys.stderr)
                times[repetition_index, param_combination_index] = np.nan

    print(f"# command template: {command_template}")
    print(f"# repetition count: {repetition_count}")
    if repetition_count > 1:
        times_mean = np.nanmean(times, axis=0)
        times_median = np.nanmedian(times, axis=0)
        times_stdev = np.nanstd(times, axis=0)
        times_min = np.nanmin(times, axis=0)
        times_max = np.nanmax(times, axis=0)
        print(f"id;"
              f"{';'.join(param_names)};"
              f"time-mean;"
              f"time-median;"
              f"time-stdev;"
              f"time-min;"
              f"time-max")
        for param_combination_index in range(param_combination_count):
            param_values = param_values_combinations[param_combination_index]
            print(f"{param_combination_index};"
                  f"{';'.join(param_values)};"
                  f"{times_mean[param_combination_index]};"
                  f"{times_median[param_combination_index]};"
                  f"{times_stdev[param_combination_index]};"
                  f"{times_min[param_combination_index]};"
                  f"{times_max[param_combination_index]}")
    else:
        print(f"id;"
              f"{';'.join(param_names)};"
              f"time")
        for param_combination_index in range(param_combination_count):
            param_values = param_values_combinations[param_combination_index]
            print(f"{param_combination_index};"
                  f"{';'.join(param_values)};"
                  f"{times[0, param_combination_index]}")


def _apply_template(template, params) -> str:
    result = template
    for param_name, param_value in params:
        result = result.replace("${" + param_name + "}", str(param_value))
    return result



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import click

from xcube.cli.common import (parse_cli_kwargs,
                              cli_option_quiet,
                              cli_option_verbosity,
                              configure_cli_output)

DEFAULT_OUTPUT_PATH = 'out.zarr'


# noinspection PyShadowingBuiltins
@click.command(name="chunk")
@click.argument('cube')
@click.option('--output', '-o', metavar='OUTPUT', default=DEFAULT_OUTPUT_PATH,
              help=f'Output path. Defaults to {DEFAULT_OUTPUT_PATH!r}')
@click.option('--format', '-f', metavar='FORMAT',
              type=click.Choice(['zarr', 'netcdf']),
              help="Format of the output. If not given, guessed from OUTPUT.")
@click.option('--params', '-p', metavar='PARAMS',
              help="Parameters specific for the output format."
                   " Comma-separated list of <key>=<value> pairs.")
@click.option('--chunks', '-C', metavar='CHUNKS', nargs=1, default=None,
              help='Chunk sizes for each dimension.'
                   ' Comma-separated list of <dim>=<size> pairs,'
                   ' e.g. "time=1,lat=270,lon=270"')
@cli_option_quiet
@cli_option_verbosity
def chunk(cube,
          output,
          format=None,
          params=None,
          chunks=None,
          quiet=None,
          verbosity=None):
    """
    (Re-)chunk xcube dataset.
    Changes the external chunking of all variables of CUBE according to CHUNKS and writes
    the result to OUTPUT.

    Note: There is a possibly more efficient way to (re-)chunk datasets through the
    dedicated tool "rechunker", see https://rechunker.readthedocs.io.
    """
    configure_cli_output(quiet=quiet, verbosity=verbosity)

    chunk_sizes = None
    if chunks:
        chunk_sizes = parse_cli_kwargs(chunks, metavar="CHUNKS")
        for k, v in chunk_sizes.items():
            if not isinstance(v, int) or v <= 0:
                raise click.ClickException("Invalid value for CHUNKS, "
                                           f"chunk sizes must be positive integers: {chunks}")

    write_kwargs = dict()
    if params:
        write_kwargs = parse_cli_kwargs(params, metavar="PARAMS")

    from xcube.core.chunk import chunk_dataset
    from xcube.core.dsio import guess_dataset_format
    from xcube.core.dsio import open_dataset, write_dataset

    format_name = format if format else guess_dataset_format(output)

    with open_dataset(input_path=cube) as ds:
        if chunk_sizes:
            for k in chunk_sizes:
                if k not in ds.dims:
                    raise click.ClickException("Invalid value for CHUNKS, "
                                               f"{k!r} is not the name of any dimension: {chunks}")

        chunked_dataset = chunk_dataset(ds, chunk_sizes=chunk_sizes, format_name=format_name)
        write_dataset(chunked_dataset, output_path=output, format_name=format_name, **write_kwargs)



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Dict, Any

import pandas as pd
import xarray as xr

from xcube.core.gridmapping import GridMapping
from xcube.version import version
from .transformer import CubeTransformer
from .transformer import TransformedCube
from ..config import CubeConfig


class CubeMetadataAdjuster(CubeTransformer):
    """Adjust a cube's metadata."""

    def transform_cube(self,
                       cube: xr.Dataset,
                       gm: GridMapping,
                       cube_config: CubeConfig) -> TransformedCube:
        history = cube.attrs.get('history')
        if isinstance(history, str):
            history = [history]
        elif isinstance(history, (list, tuple)):
            history = list(history)
        else:
            history = []
        history.append(
            dict(
                program=f'xcube gen2, version {version}',
                cube_config=cube_config.to_dict(),
            )
        )
        cube = cube.assign_attrs(
            Conventions='CF-1.7',
            history=history,
            date_created=pd.Timestamp.now().isoformat(),
            # TODO: adjust temporal metadata too
            **get_geospatial_attrs(gm)
        )
        if cube_config.metadata:
            self._check_for_self_destruction(cube_config.metadata)
            cube.attrs.update(cube_config.metadata)
        if cube_config.variable_metadata:
            for var_name, metadata in cube_config.variable_metadata.items():
                if var_name in cube.variables and metadata:
                    cube[var_name].attrs.update(metadata)
        return cube, gm, cube_config

    @staticmethod
    def _check_for_self_destruction(metadata: Dict[str, Any]):
        key = 'inverse_fine_structure_constant'
        value = 137
        if key in metadata and metadata[key] != value:
            # Note, this is an easter egg that causes
            # an intended internal error for testing
            raise ValueError(f'{key} must be {value}'
                             f' or running in wrong universe')


def get_geospatial_attrs(gm: GridMapping) -> Dict[str, Any]:
    return dict(gm.to_dataset_attrs())



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import traceback
from abc import ABC, abstractmethod
from typing import Optional, TypeVar, Type

from xcube.core.store import DataStorePool
from xcube.core.store import DataStorePoolLike
from xcube.util.assertions import assert_instance
from xcube.util.assertions import assert_true
from .error import CubeGeneratorError
from .progress import ConsoleProgressObserver
from .remote.config import ServiceConfigLike
from .request import CubeGeneratorRequestLike
from .response import CubeGeneratorResult
from .response import CubeInfoResult
from .response import GenericCubeGeneratorResult

R = TypeVar('R', bound=GenericCubeGeneratorResult)


class CubeGenerator(ABC):
    """
    Abstract base class for cube generators.

    Use the ``CubeGenerator.load()`` method to instantiate new
    cube generators.
    """

    @classmethod
    def new(cls,
            service_config: Optional[ServiceConfigLike] = None,
            stores_config: Optional[DataStorePoolLike] = None,
            raise_on_error: bool = False,
            verbosity: int = 0,
            **kwargs) -> 'CubeGenerator':
        """
        Create a new cube generator from given configurations.

        If *service_config* is given, it describes a remote xcube
        generator remote, otherwise a local cube generator is configured
        using optional *stores_config*.

        The *service_config* parameter can be passed in different ways:

        * An instance of :class:ServiceConfig.
        * A ``str``. Then it is interpreted as a path to a YAML or JSON file
          and the remote configuration is loaded from this file.
          The file content may include template variables that are
          interpolated by environment variables,
          e.g. "${XCUBE_GEN_CLIENT_SECRET}".
        * A ``dict``. Then it is interpreted as a remote configuration
          JSON object.

        If *stores_config* is given, it describes a pool of data stores to be
        used as input and output for the cube generator. *stores_config*
        if a mapping of store instance identifiers to configured store
        instances. A store instance is a dictionary that has a mandatory
        "store_id" property which is a name of a registered xcube data store.
        as well as an optional "store_params" property that may define data
        store specific parameters.

        Similar to *service_config*, the *stores_config* parameter
        can be passed in different ways:

        * An instance of :class:DataStorePool.
        * A ``str``. Then it is interpreted as a YAML or JSON file path
          and the stores configuration is loaded from this file.
        * A ``dict``. Then it is interpreted as a stores configuration
          JSON object.

        The *service_config* and *stores_config* parameters cannot
        be given both.

        :param service_config: Service configuration.
        :param stores_config: Data stores configuration.
        :param raise_on_error: Whether to raise a CubeGeneratorError
            exception on generator failures. If False, the default,
            the returned result will have the "status" field set to "error"
            while other fields such as "message", "traceback", "output"
            provide more failure details.
        :param verbosity: Level of verbosity, 0 means off.
        :param kwargs: Extra arguments passed to the generator constructors.
        """
        if service_config is not None:
            from .remote.config import ServiceConfig
            from .remote.generator import RemoteCubeGenerator
            assert_true(stores_config is None,
                        'service_config and stores_config cannot be'
                        ' given at the same time.')
            assert_instance(service_config,
                            (str, dict, ServiceConfig, type(None)),
                            'service_config')
            service_config = ServiceConfig.normalize(service_config) \
                if service_config is not None else None
            return RemoteCubeGenerator(service_config=service_config,
                                       raise_on_error=raise_on_error,
                                       verbosity=verbosity,
                                       **kwargs)
        else:
            from .local.generator import LocalCubeGenerator
            assert_instance(stores_config,
                            (str, dict, DataStorePool, type(None)),
                            'stores_config')
            store_pool = DataStorePool.normalize(stores_config) \
                if stores_config is not None else None
            return LocalCubeGenerator(store_pool=store_pool,
                                      raise_on_error=raise_on_error,
                                      verbosity=verbosity)

    def __init__(self,
                 raise_on_error: bool = False,
                 verbosity: int = 0):
        self._raise_on_error = raise_on_error
        self._verbosity = verbosity

    def get_cube_info(self, request: CubeGeneratorRequestLike) \
            -> CubeInfoResult:
        """
        Get data cube information for given *request*.

        The *request* argument can be
        * an instance of ``CubeGeneratorRequest``;
        * a ``dict``. In this case it is interpreted as JSON object and
          parsed into a ``CubeGeneratorRequest``;
        * a ``str``. In this case it is interpreted as path to a
          YAML or JSON file, which is loaded and
          parsed into a ``CubeGeneratorRequest``.

        :param request: Cube generator request.
        :return: a cube information result
            of type :class:CubeInfoResult
        :raises CubeGeneratorError: if cube info generation failed
        :raises DataStoreError: if data store access failed
        """
        try:
            result = self._get_cube_info(request)
        except CubeGeneratorError as e:
            if self._raise_on_error:
                raise e
            return self._new_cube_generator_error_result(
                CubeInfoResult, e
            )
        if result.status == 'error':
            if self._raise_on_error:
                raise self._new_generator_error_from_result(result)
        return result

    def generate_cube(self, request: CubeGeneratorRequestLike) \
            -> CubeGeneratorResult:
        """
        Generate the data cube for given *request*.

        The *request* argument can be
        * an instance of ``CubeGeneratorRequest``;
        * a ``dict``. In this case it is interpreted as JSON object and
          parsed into a ``CubeGeneratorRequest``;
        * a ``str``. In this case it is interpreted as path to a
          YAML or JSON file, which is loaded and
          parsed into a ``CubeGeneratorRequest``.

        Returns the cube reference which can be used as ``data_id`` in
        ``store.open_data(data_id)`` where *store*  refers to the
        store configured in ``output_config`` of the cube generator request.

        :param request: Cube generator request.
        :return: the cube generation result
            of type :class:CubeGeneratorResult
        :raises CubeGeneratorError: if cube generation failed
        :raises DataStoreError: if data store access failed
        """

        if self._verbosity:
            ConsoleProgressObserver().activate()

        try:
            result = self._generate_cube(request)
        except CubeGeneratorError as e:
            if self._raise_on_error:
                raise e
            return self._new_cube_generator_error_result(
                CubeGeneratorResult, e
            )
        finally:
            if self._verbosity:
                ConsoleProgressObserver().deactivate()

        if result.status == 'error':
            if self._raise_on_error:
                raise self._new_generator_error_from_result(result)
        return result

    @abstractmethod
    def _get_cube_info(self, request: CubeGeneratorRequestLike) \
            -> CubeInfoResult:
        """
        The implementation of the :meth:`get_cube_info` method
        """

    @abstractmethod
    def _generate_cube(self, request: CubeGeneratorRequestLike) \
            -> CubeGeneratorResult:
        """
        The implementation of the :meth:`generate_cube` method
        """

    @classmethod
    def _new_cube_generator_error_result(
            cls,
            result_type: Type[R],
            e: CubeGeneratorError
    ) -> R:
        tb = e.remote_traceback
        if tb is None and e.__traceback__ is not None:
            tb = traceback.format_tb(e.__traceback__)
        return result_type(status='error',
                           message=f'{e}',
                           status_code=e.status_code,
                           output=e.remote_output,
                           traceback=tb)

    @classmethod
    def _new_generator_error_from_result(cls,
                                         result: GenericCubeGeneratorResult):
        return CubeGeneratorError(result.message,
                                  status_code=result.status_code,
                                  remote_output=result.output,
                                  remote_traceback=result.traceback)



import threading
from typing import Dict, List, Mapping, Any, Union, Sequence, Optional
from deprecated import deprecated
import numpy as np
import pandas as pd
import xarray as xr

from xcube.core.chunk import chunk_dataset
from xcube.core.dsio import open_cube, write_cube
from xcube.core.dump import dump_dataset
from xcube.core.extract import DEFAULT_INDEX_NAME_PATTERN
from xcube.core.extract import get_cube_point_indexes
from xcube.core.extract import get_cube_values_for_indexes
from xcube.core.extract import get_cube_values_for_points
from xcube.core.extract import get_dataset_indexes
from xcube.core.gridmapping import GridMapping
from xcube.core.level import compute_levels
from xcube.core.new import new_cube
from xcube.core.normalize import DatasetIsNotACubeError
from xcube.core.normalize import decode_cube
from xcube.core.schema import CubeSchema, get_cube_schema
from xcube.core.select import select_variables_subset
from xcube.core.vars2dim import vars_to_dim
from xcube.core.verify import verify_cube


@xr.register_dataset_accessor('xcube')
class DatasetAccessor:
    """
    The xcube xarray API.

    The API is made available via the ``xcube`` attribute of
    xarray.Dataset instances.

    It defines new xcube-specific properties for xarray datasets:

    * :attr:cube The subset of variables of this dataset
        which all have cube dimensions (time, ..., <y_name>, <x_name>).
        May be an empty dataset.
    * :attr:non_cube The subset of variables of this dataset
        minus the data variables from :attr:cube.
        May be the same as this dataset.
    * :attr:gm The grid mapping used by this dataset.
        It is an instance of :class:GridMapping.
        May be None, if this dataset does not define a grid mapping.
    """

    def __init__(self, dataset: xr.Dataset):
        self._dataset: xr.Dataset = dataset
        self._cube_subset: Optional[xr.Dataset] = None
        self._grid_mapping: Optional[GridMapping] = None
        self._lock = threading.RLock()

    @property
    def cube(self) -> xr.Dataset:
        if self._cube_subset is None:
            with self._lock:
                self._init_cube_subset()
        return self._cube_subset

    @property
    def non_cube(self) -> xr.Dataset:
        if self._cube_subset is None:
            with self._lock:
                self._init_cube_subset()
        return self._non_cube_subset

    @property
    def gm(self) -> Optional[GridMapping]:
        if self._cube_subset is None:
            with self._lock:
                self._init_cube_subset()
        return self._grid_mapping

    def _init_cube_subset(self):
        try:
            cube, grid_mapping, non_cube = decode_cube(self._dataset,
                                                       normalize=True,
                                                       force_copy=True)
        except DatasetIsNotACubeError:
            cube, grid_mapping, non_cube = xr.Dataset(), None, self._dataset
        self._cube_subset = cube
        self._grid_mapping = grid_mapping
        self._non_cube_subset = non_cube

    ########################################################################
    # Old API from here on.
    #
    # Let's quickly agree, if we should deprecate all this stuff. I guess,
    # no one uses it.
    #
    # We should only add props and methods to this accessor
    # that require a certain state to be hold.
    # Such state could be props that are expensive to recompute,
    # such as grid mappings.
    #
    # It causes too much overhead and maintenance work
    # if we continue putting any xcube function here.
    ########################################################################

    @classmethod
    def new(cls, **kwargs) -> xr.Dataset:
        """
        Create a new empty cube. Useful for testing.

        Refer to :func:`xcube.core.new.new_cube` for details.
        """
        return new_cube(**kwargs)

    @classmethod
    def open(cls, input_path: str, format_name: str = None, **kwargs) -> xr.Dataset:
        """
        The ``read`` method as context manager that auto-closes the data cube read.

        :param input_path: input path
        :param format_name: format, e.g. "zarr" or "netcdf4"
        :param kwargs: format-specific keyword arguments
        :return: dataset object
        """
        return open_cube(input_path, format_name=format_name, **kwargs)

    def write(self, output_path: str, format_name: str = None, **kwargs) -> xr.Dataset:
        """
        Write this cube to *output_path*.
        If *format* is not provided it will be guessed from *output_path*.

        :param output_path: output path
        :param format_name: format, e.g. "zarr" or "netcdf4"
        :param kwargs: format-specific keyword arguments
        :return: the input dataset
        """
        return write_cube(self._dataset, output_path, format_name=format_name, **kwargs)

    def values_for_points(self,
                          points: Union[xr.Dataset, pd.DataFrame, Mapping[str, Any]],
                          var_names: Sequence[str] = None,
                          index_name_pattern: str = DEFAULT_INDEX_NAME_PATTERN,
                          include_indexes: bool = False,
                          method: str = "nearest",
                          cube_asserted: bool = False):
        """
        Extract values from cube variables at given coordinates in *points*.

        :param points: Dictionary that maps dimension name to coordinate arrays.
        :param var_names: An optional list of names of data variables in *cube* whose values shall be extracted.
        :param index_name_pattern: A naming pattern for the computed indexes columns.
               Must include "{name}" which will be replaced by the dimension name.
        :param include_indexes: Whether to include computed indexes in return value.
        :param method: "nearest" or "linear".
        :param cube_asserted: If False, *cube* will be verified, otherwise it is expected to be a valid cube.
        :return: A new data frame whose columns are values from *cube* variables at given *points*.
        """
        return get_cube_values_for_points(self._dataset,
                                          points,
                                          var_names=var_names,
                                          index_name_pattern=index_name_pattern,
                                          include_indexes=include_indexes,
                                          method=method,
                                          cube_asserted=cube_asserted)

    def values_for_indexes(self,
                           indexes: Union[xr.Dataset, pd.DataFrame, Mapping[str, Any]],
                           var_names: Sequence[str] = None,
                           index_name_pattern: str = DEFAULT_INDEX_NAME_PATTERN,
                           method: str = "nearest",
                           cube_asserted: bool = False) -> xr.Dataset:
        """
        Get values from this cube at given *indexes*.

        :param indexes: A mapping from column names to index and fraction arrays for all cube dimensions.
        :param var_names: An optional list of names of data variables in *cube* whose values shall be extracted.
        :param index_name_pattern: A naming pattern for the computed indexes columns.
               Must include "{name}" which will be replaced by the dimension name.
        :param method: "nearest" or "linear".
        :param cube_asserted: If False, *cube* will be verified, otherwise it is expected to be a valid cube.
        :return: A new data frame whose columns are values from *cube* variables at given *indexes*.
        """
        return get_cube_values_for_indexes(self._dataset,
                                           indexes,
                                           data_var_names=var_names,
                                           index_name_pattern=index_name_pattern,
                                           method=method,
                                           cube_asserted=cube_asserted)

    def point_indexes(self,
                      points: Union[xr.Dataset, pd.DataFrame, Mapping[str, Any]],
                      dim_name_mapping: Mapping[str, str] = None,
                      index_name_pattern: str = DEFAULT_INDEX_NAME_PATTERN,
                      index_dtype=np.float64,
                      cube_asserted: bool = False):
        """
        Get indexes of given coordinates in *points* into this cube.

        :param points: A mapping from column names to column data arrays, which must all have the same length.
        :param dim_name_mapping: A mapping from dimension names in *cube* to column names in *points*.
        :param index_name_pattern: A naming pattern for the computed indexes columns.
               Must include "{name}" which will be replaced by the dimension name.
        :param index_dtype: Numpy data type for the indexes. If it is a floating point type (default),
               then *indexes* will contain fractions, which may be used for interpolation.
               For out-of-range coordinates in *points*, indexes will be -1 if *index_dtype* is an integer type, and NaN,
               if *index_dtype* is a floating point types.
        :param cube_asserted: If False, *cube* will be verified, otherwise it is expected to be a valid cube.
        :return: A dataset containing the index columns.
        """
        return get_cube_point_indexes(self._dataset,
                                      points,
                                      dim_name_mapping=dim_name_mapping,
                                      index_name_pattern=index_name_pattern,
                                      index_dtype=index_dtype,
                                      cube_asserted=cube_asserted)

    def indexes(self,
                coord_var_name: str,
                coord_values: Union[xr.DataArray, np.ndarray],
                index_dtype=np.float64) -> Union[xr.DataArray, np.ndarray]:
        """
        Compute the indexes into a coordinate variable *coord_var_name* of this cube
        for the given coordinate values *coord_values*.

        The coordinate variable's labels must be monotonic increasing or decreasing,
        otherwise the result will be nonsense.

        For any value in *coord_values* that is out of the bounds of the coordinate variable's values,
        the index depends on the value of *index_dtype*. If *index_dtype* is an integer type, invalid indexes are
        encoded as -1 while for floating point types, NaN will be used.

        Returns the indexes as an array-like object of type *dtype*.

        :param coord_var_name: Name of a coordinate variable.
        :param coord_values: Array-like coordinate values.
        :param index_dtype: Numpy data type for the indexes. If it is a floating point type (default),
               then *indexes* will contain fractions, which may be used for interpolation.
               For out-of-range coordinates in *points*, indexes will be -1 if *index_dtype* is an integer type, and NaN,
               if *index_dtype* is a floating point types.
        :return: The indexes and their fractions as a tuple of numpy int64 and float64 arrays.
        """
        return get_dataset_indexes(self._dataset,
                                   coord_var_name=coord_var_name,
                                   coord_values=coord_values,
                                   index_dtype=index_dtype)

    def chunk(self, chunk_sizes: Dict[str, int] = None, format_name: str = None) -> xr.Dataset:
        """
        Chunk this dataset and update encodings for given format.

        :param chunk_sizes: mapping from dimension name to new chunk size
        :param format_name: format, e.g. "zarr" or "netcdf4"
        :return: the re-chunked dataset
        """
        return chunk_dataset(self._dataset,
                             chunk_sizes=chunk_sizes,
                             format_name=format_name)

    def vars_to_dim(self, dim_name: str = 'newdim'):
        """
        Convert data variables into a dimension

        :param dim_name: The name of the new dimension ['vars']
        :return: A new xcube dataset with the new dimension.
        """
        return vars_to_dim(self._dataset, dim_name)

    def dump(self,
             var_names=None,
             show_var_encoding=False) -> str:
        """
        Dump this dataset or its variables into a text string.

        :param var_names: names of variables to be dumped
        :param show_var_encoding: also dump variable encodings?
        :return: the dataset dump
        """
        return dump_dataset(self._dataset,
                            var_names=var_names,
                            show_var_encoding=show_var_encoding)

    def verify(self) -> List[str]:
        """
        Verify that this dataset is a valid xcube dataset.

        Returns a list of issues, which is empty if this dataset is a valid xcube dataset.

        :return: List of issues or empty list.
        """
        return verify_cube(self._dataset)

    def select_variables_subset(self, var_names: Sequence[str] = None):
        """
        Select data variable from given *dataset* and create new dataset.

        :param var_names: The names of data variables to select.
        :return: A new dataset. It is empty, if *var_names* is empty. It is *dataset*, if *var_names* is None.
        """
        return select_variables_subset(self._dataset, var_names)

    @deprecated(version='0.13.0',
                reason='multi-level datasets should be represented by'
                       ' class xcube.core.mldataset.MultiLevelDataset')
    def levels(self, **kwargs) -> List[xr.Dataset]:
        """
        Transform this dataset into the levels of a multi-level pyramid with spatial resolution
        decreasing by a factor of two in both spatial dimensions.

        It is assumed that the spatial dimensions of each variable are the inner-most, that is, the last two elements
        of a variable's shape provide the spatial dimension sizes.

        :param spatial_dims: If given, only variables are considered whose last to dimension elements match the given *spatial_dims*.
        :param spatial_shape: If given, only variables are considered whose last to shape elements match the given *spatial_shape*.
        :param spatial_tile_shape: If given, chunking will match the provided *spatial_tile_shape*.
        :param var_names: Variables to consider. If None, all variables with at least two dimensions are considered.
        :param max_num_levels: If given, the maximum number of pyramid levels.
        :param post_process_level: If given, the function will be called for each level and must return a dataset.
        :param progress_monitor: If given, the function will be called for each level.
        :return: A list of dataset instances representing the multi-level pyramid.
        """
        return compute_levels(self._dataset, **kwargs)

    @property
    def schema(self) -> CubeSchema:
        return get_cube_schema(self._dataset)



# The MIT License (MIT)
# Copyright (c) 2020 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from xcube.constants import EXTENSION_POINT_CLI_COMMANDS
from xcube.constants import EXTENSION_POINT_DATASET_IOS
from xcube.constants import EXTENSION_POINT_DATA_OPENERS
from xcube.constants import EXTENSION_POINT_DATA_STORES
from xcube.constants import EXTENSION_POINT_DATA_WRITERS
from xcube.constants import EXTENSION_POINT_INPUT_PROCESSORS
from xcube.constants import EXTENSION_POINT_SERVER_APIS
from xcube.constants import EXTENSION_POINT_SERVER_FRAMEWORKS
from xcube.constants import FORMAT_NAME_CSV
from xcube.constants import FORMAT_NAME_MEM
from xcube.constants import FORMAT_NAME_NETCDF4
from xcube.constants import FORMAT_NAME_ZARR
from xcube.util import extension


def init_plugin(ext_registry: extension.ExtensionRegistry):
    """
    Register xcube's standard extensions.
    """
    _register_input_processors(ext_registry)
    _register_dataset_ios(ext_registry)
    _register_cli_commands(ext_registry)
    _register_data_stores(ext_registry)
    _register_data_accessors(ext_registry)
    _register_server_apis(ext_registry)
    _register_server_frameworks(ext_registry)


def _register_input_processors(ext_registry: extension.ExtensionRegistry):
    """
    Register xcube's standard input processors used by "xcube gen" and gen_cube().
    """
    ext_registry.add_extension(
        loader=extension.import_component('xcube.core.gen.iproc:DefaultInputProcessor'),
        point=EXTENSION_POINT_INPUT_PROCESSORS, name='default',
        description='Single-scene NetCDF/CF inputs in xcube standard format'
    )


def _register_dataset_ios(ext_registry: extension.ExtensionRegistry):
    """
    Register xcube's standard dataset I/O components used by various CLI and API functions.
    """
    ext_registry.add_extension(
        loader=extension.import_component('xcube.core.dsio:ZarrDatasetIO', call=True),
        point=EXTENSION_POINT_DATASET_IOS, name=FORMAT_NAME_ZARR,
        description='Zarr file format (http://zarr.readthedocs.io)',
        ext='zarr', modes={'r', 'w', 'a'}
    )
    ext_registry.add_extension(
        loader=extension.import_component('xcube.core.dsio:Netcdf4DatasetIO',
                                          call=True),
        point=EXTENSION_POINT_DATASET_IOS, name=FORMAT_NAME_NETCDF4,
        description='NetCDF-4 file format',
        ext='nc', modes={'r', 'w', 'a'}
    )
    ext_registry.add_extension(
        loader=extension.import_component('xcube.core.dsio:CsvDatasetIO', call=True),
        point=EXTENSION_POINT_DATASET_IOS, name=FORMAT_NAME_CSV,
        description='CSV file format',
        ext='csv', modes={'r', 'w'}
    )
    ext_registry.add_extension(
        loader=extension.import_component('xcube.core.dsio:MemDatasetIO', call=True),
        point=EXTENSION_POINT_DATASET_IOS, name=FORMAT_NAME_MEM,
        description='In-memory dataset I/O',
        ext='mem', modes={'r', 'w', 'a'}
    )


_FS_STORAGE_ITEMS = (
    ('file', 'local filesystem'),
    ('s3', 'AWS S3 compatible object storage'),
    ('abfs', 'Azure blob compatible object storage'),
    ('memory', 'in-memory filesystem'),
    ('ftp', 'FTP filesystem'),
    ('reference', 'reference filesystem'),
)

_FS_DATA_ACCESSOR_ITEMS = (
    ('dataset', 'netcdf',
     'xarray.Dataset in NetCDF format'),
    ('dataset', 'zarr',
     'xarray.Dataset in Zarr format'),
    ('dataset', 'levels',
     'xarray.Dataset in leveled Zarr format'),
    ('mldataset', 'levels',
     'xcube.core.mldataset.MultiLevelDataset in leveled Zarr format'),
    ('dataset', 'geotiff',
     'xarray.Dataset in GeoTIFF or COG format'),
    ('mldataset', 'geotiff',
     'xcube.core.mldataset.MultiLevelDataset in GeoTIFF or COG format'),
    ('geodataframe', 'shapefile',
     'gpd.GeoDataFrame in ESRI Shapefile format'),
    ('geodataframe', 'geojson',
     'gpd.GeoDataFrame in GeoJSON format'),
)

_FS_DATA_OPENER_ITEMS = _FS_DATA_ACCESSOR_ITEMS
_FS_DATA_WRITER_ITEMS = _FS_DATA_ACCESSOR_ITEMS


def _register_data_stores(ext_registry: extension.ExtensionRegistry):
    """
    Register xcube's standard data stores.
    """
    fs_ds_cls_factory = 'xcube.core.store.fs.registry:get_fs_data_store_class'
    for storage_id, storage_description in _FS_STORAGE_ITEMS:
        fs_ds_cls_loader = extension.import_component(fs_ds_cls_factory,
                                                      call_args=[storage_id])
        ext_registry.add_extension(
            point=EXTENSION_POINT_DATA_STORES,
            loader=fs_ds_cls_loader,
            name=storage_id,
            description=f'Data store that uses a {storage_description}'
        )

    ref_ds_cls = 'xcube.core.store.ref.store:ReferenceDataStore'
    ref_ds_cls_loader = extension.import_component(ref_ds_cls)
    ext_registry.add_extension(
        point=EXTENSION_POINT_DATA_STORES,
        loader=ref_ds_cls_loader,
        name="reference",
        description=f'Data store that uses Kerchunk references'
    )


def _register_data_accessors(ext_registry: extension.ExtensionRegistry):
    """
    Register xcube's standard data accessors.
    """
    factory = 'xcube.core.store.fs.registry:get_fs_data_accessor_class'

    # noinspection PyShadowingNames
    def _add_fs_data_accessor_ext(point: str,
                                  ext_type: str,
                                  protocol: str,
                                  data_type: str,
                                  format_id: str):
        factory_args = (protocol, data_type, format_id)
        loader = extension.import_component(factory,
                                            call_args=factory_args)
        ext_registry.add_extension(
            point=point,
            loader=loader,
            name=f'{data_type}:{format_id}:{protocol}',
            description=f'Data {ext_type} for'
                        f' a {data_accessor_description}'
                        f' in {storage_description}'
        )

    for protocol, storage_description in _FS_STORAGE_ITEMS:
        for data_type, format_id, data_accessor_description \
                in _FS_DATA_OPENER_ITEMS:
            _add_fs_data_accessor_ext(EXTENSION_POINT_DATA_OPENERS,
                                      'opener',
                                      protocol,
                                      data_type,
                                      format_id)
        for data_type, format_id, data_accessor_description \
                in _FS_DATA_WRITER_ITEMS:
            _add_fs_data_accessor_ext(EXTENSION_POINT_DATA_WRITERS,
                                      'writer',
                                      protocol,
                                      data_type,
                                      format_id)


def _register_cli_commands(ext_registry: extension.ExtensionRegistry):
    """
    Register xcube's standard CLI commands.
    """

    cli_command_names = [
        'chunk',
        'compute',
        'benchmark',
        'dump',
        'extract',
        'gen',
        'gen2',
        'genpts',
        'grid',
        'level',
        'optimize',
        'patch',
        'prune',
        'rectify',
        'resample',
        'serve',
        'vars2dim',
        'verify',
        'versions',

        # Experimental + Hidden
        'io',
    ]

    for cli_command_name in cli_command_names:
        ext_registry.add_extension(
            loader=extension.import_component(
                f'xcube.cli.{cli_command_name}:{cli_command_name}'),
            point=EXTENSION_POINT_CLI_COMMANDS,
            name=cli_command_name
        )


def _register_server_apis(ext_registry: extension.ExtensionRegistry):
    """
    Register xcube's standard server APIs.
    """
    server_api_names = [
        'meta',
        'auth',
        'compute',
        'places',
        'styles',
        'datasets',
        'tiles',
        'timeseries',
        'volumes',
        'ows.coverages',
        'ows.stac',
        'ows.wmts',
        's3',
        'viewer',
    ]
    for api_name in server_api_names:
        ext_registry.add_extension(
            loader=extension.import_component(
                f'xcube.webapi.{api_name}:api'
            ),
            point=EXTENSION_POINT_SERVER_APIS,
            name=api_name
        )


def _register_server_frameworks(ext_registry: extension.ExtensionRegistry):
    server_framework_names = [
        'tornado',
        'flask',
    ]
    for framework_name in server_framework_names:
        ext_registry.add_extension(
            loader=extension.import_component(
                f'xcube.server.webservers.{framework_name}'
                f':{framework_name.capitalize()}Framework',
            ),
            point=EXTENSION_POINT_SERVER_FRAMEWORKS,
            name=framework_name
        )



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import collections.abc
from typing import Union, Iterator, Tuple, Mapping

import xarray as xr
import zarr.storage

from xcube.core.mldataset import MultiLevelDataset
from xcube.server.api import ApiError


class ObjectStorage(collections.abc.Mapping):
    """Represents the emulated object storage for the S3 API.

    Keys are strings and expected to have format

    * "{dataset_id}/{level}.zarr/{*path}" for multi-level datasets and
    * "{dataset_id}/{*path}" for other datasets.

    :param datasets: Mapping from dataset Identifier
        to (multi-level) datasets.
    """

    def __init__(self,
                 datasets: Mapping[
                     str,
                     Union[xr.Dataset, MultiLevelDataset]
                 ]):
        self.datasets = datasets

    def __len__(self) -> int:
        return sum(1 for _ in iter(self))

    def __iter__(self) -> Iterator[str]:
        for dataset_id, dataset in self.datasets.items():
            if isinstance(dataset, MultiLevelDataset):
                for level in range(dataset.num_levels):
                    level_dataset = dataset.get_dataset(level)
                    zarr_store = level_dataset.zarr_store.get()
                    for k in zarr_store.keys():
                        yield f"{dataset_id}/{level}.zarr/{k}"

            else:
                zarr_store = dataset.zarr_store.get()
                for k in zarr_store.keys():
                    yield f"{dataset_id}/{k}"

    def __contains__(self, key: str) -> bool:
        """Overridden to avoid a call to __getitem__(),
        which will load data, but we want this to happen
        for direct __getitem__() calls only."""
        try:
            zarr_store, item_key = self._parse_key(key)
        except (KeyError, ApiError.NotFound):
            # Parsing failed, must be invalid key
            return False
        # Now check item_key
        return item_key in zarr_store

    def __getitem__(self, key: str) -> bytes:
        """Get bytes object for *key*."""
        zarr_store, item_key = self._parse_key(key)
        value = zarr_store[item_key]
        if not isinstance(value, bytes):
            raise RuntimeError(
                f"Zarr store of type {type(zarr_store).__name__}"
                f" must return type bytes for key {key!r},"
                f" but was {type(value).__name__}"
            )
        return value

    def _parse_key(self, key: str) -> Tuple[zarr.storage.BaseStore, str]:
        """Parses a given *key* which is expected to have format
        "{dataset_id}/{level}.zarr/{*path}" for multi-level datasets and
        "{dataset_id}/{*path}" for other datasets.
        """
        try:
            dataset_id, item_key = key.split("/", maxsplit=1)
        except ValueError:
            raise KeyError(key)

        try:
            dataset = self.datasets[dataset_id]
        except ApiError as e:
            raise KeyError(key) from e

        if isinstance(dataset, MultiLevelDataset):
            try:
                level_dataset_id, item_key = item_key.split("/", maxsplit=1)
                level_dataset_id, _ = level_dataset_id.split(".", maxsplit=1)
                level = int(level_dataset_id)
            except ValueError:
                raise KeyError(key)
            if not (0 <= level < dataset.num_levels):
                raise KeyError(key)
            level_dataset = dataset.get_dataset(level)
            zarr_store = level_dataset.zarr_store.get()
        else:
            zarr_store = dataset.zarr_store.get()

        return zarr_store, item_key



# The MIT License (MIT)
# Copyright (c) 2019 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import List

import numpy as np
import xarray as xr

from xcube.core.geom import is_lon_lat_dataset
from xcube.core.schema import get_dataset_xy_var_names, get_dataset_time_var_name


def assert_cube(dataset: xr.Dataset, name=None) -> xr.Dataset:
    """
    Assert that the given *dataset* is a valid xcube dataset.

    :param dataset: The dataset to be validated.
    :param name: Optional parameter name.
    :raise: ValueError, if dataset is not a valid xcube dataset
    """
    report = verify_cube(dataset)
    if report:
        message = f"Dataset" + (name + " " if name else " ")
        message += "is not a valid xcube dataset, because:\n"
        message += "- " + ";\n- ".join(report) + "."
        raise ValueError(message)

    return dataset


def verify_cube(dataset: xr.Dataset) -> List[str]:
    """
    Verify the given *dataset* for being a valid xcube dataset.

    The tool verifies that *dataset*
    * defines two spatial x,y coordinate variables, that are 1D, non-empty, using correct units;
    * defines a time coordinate variables, that are 1D, non-empty, using correct units;
    * has valid bounds variables for spatial x,y and time coordinate variables, if any;
    * has any data variables and that they are valid, e.g. min. 3-D, all have
      same dimensions, have at least the dimensions dim(time), dim(y), dim(x) in that order.

    Returns a list of issues, which is empty if *dataset* is a valid xcube dataset.

    :param dataset: A dataset to be verified.
    :return: List of issues or empty list.
    """
    report = []

    xy_var_names = get_dataset_xy_var_names(dataset, must_exist=False)
    if xy_var_names is None:
        report.append(f"missing spatial x,y coordinate variables")

    time_var_name = get_dataset_time_var_name(dataset, must_exist=False)
    if time_var_name is None:
        report.append(f"missing time coordinate variable")

    if time_var_name:
        _check_time(dataset, time_var_name, report)
    if xy_var_names and is_lon_lat_dataset(dataset, xy_var_names=xy_var_names):
        _check_lon_or_lat(dataset, xy_var_names[0], -180., 180., report)
        _check_lon_or_lat(dataset, xy_var_names[1], -90., 90., report)

    if xy_var_names and time_var_name:
        _check_data_variables(dataset, xy_var_names, time_var_name, report)

    if xy_var_names:
        _check_coord_equidistance(dataset, xy_var_names[0], xy_var_names[0], report)
        _check_coord_equidistance(dataset, xy_var_names[1], xy_var_names[1], report)

    return report


def _check_coord_equidistance(dataset, coord_name, dim_name, report, rtol=None):
    diff = dataset[coord_name].diff(dim=dim_name)
    if not _check_equidistance_from_diff(dataset, diff, rtol=rtol):
        report.append(f"coordinate variable {coord_name!r} is not equidistant")

    bnds_name = dataset.attrs.get('bounds', f'{coord_name}_bnds')

    if bnds_name in dataset.coords:
        diff = dataset[bnds_name].diff(dim=dim_name)
        if not _check_equidistance_from_diff(dataset, diff[:, 0], rtol=rtol):
            report.append(f"coordinate variable {bnds_name!r} is not equidistant")
        elif not _check_equidistance_from_diff(dataset, diff[:, 1], rtol=rtol):
            report.append(f"coordinate variable {bnds_name!r} is not equidistant")


def _check_equidistance_from_diff(dataset, diff, rtol=None):
    if rtol is None:
        rtol = np.abs(np.divide(diff[0], 100.00)).values

    # Check whether the bounding box intersect with the anti-meridian for lon/lat datasets.
    # This is the case when exactly one difference is negative.
    if is_lon_lat_dataset(dataset):
        ct_neg = diff.where(diff < 0).count().values
        if ct_neg == 1:
            # If a bounding box intersects with the anti-meridian, compute its correct width
            diff = xr.where(diff < 0, diff + 360., diff)
    return np.allclose(diff[0], diff, rtol=rtol)


def _check_data_variables(dataset, xy_var_names, time_var_name, report):
    x_var_name, y_var_name = xy_var_names
    x_var, y_var, time_var = dataset[x_var_name], dataset[y_var_name], dataset[time_var_name]
    x_dim, y_dim, time_dim = x_var.dims[0], y_var.dims[0], time_var.dims[0]

    first_var = None
    first_dims = None
    first_chunks = None
    for var_name, var in dataset.data_vars.items():
        dims = var.dims
        chunks = var.data.chunks if hasattr(var.data, "chunks") else None

        if ("grid_mapping_name" in var.attrs
                or 'geographic_crs_name' in var.attrs
                or "crs_wkt" in var.attrs):
            # potential CRS / grid mapping variable
            continue

        if len(dims) < 3 or dims[0] != time_dim or dims[-2] != y_dim or dims[-1] != x_dim:
            report.append(f"dimensions of data variable {var_name!r}"
                          f" must be ({time_dim!r}, ..., {y_dim!r}, {x_dim!r}),"
                          f" but were {dims!r} for {var_name!r}")

        if first_var is None:
            first_var = var
            first_dims = dims
            first_chunks = chunks
            continue

        if first_dims != dims:
            report.append("dimensions of all data variables must be same,"
                          f" but found {first_dims!r} for {first_var.name!r} "
                          f"and {dims!r} for {var_name!r}")

        if first_chunks != chunks:
            report.append("all data variables must have same chunk sizes,"
                          f" but found {first_chunks!r} for {first_var.name!r} "
                          f"and {chunks!r} for {var_name!r}")


def _check_dim(dataset, name, report):
    if name not in dataset.dims:
        report.append(f"missing dimension {name!r}")
    elif dataset.dims[name] < 0:
        report.append(f"size of dimension {name!r} must be a positive integer")


def _check_coord_var(dataset, var_name, report):
    if var_name not in dataset.coords:
        report.append(f"missing coordinate variable {var_name!r}")
        return None

    var = dataset.coords[var_name]
    if var.dims != (var_name,):
        report.append(f"coordinate variable {var_name!r} must have a single dimension {var_name!r}")
        return None

    if var.size == 0:
        report.append(f"coordinate variable {var_name!r} must not be empty")
        return None

    bnds_name = var.attrs.get('bounds', f'{var_name}_bnds')
    if bnds_name in dataset.coords:
        bnds_var = dataset.coords[bnds_name]
        expected_shape = var.size, 2
        expected_dtype = var.dtype
        if len(bnds_var.dims) != 2 or bnds_var.dims[0] != var_name:
            report.append(f"bounds coordinate variable {bnds_name!r}"
                          f" must have dimensions ({var_name!r}, <bounds_dim>)")
        if bnds_var.shape != expected_shape:
            report.append(
                f"shape of bounds coordinate variable {bnds_name!r}"
                f" must be {expected_shape!r} but was {bnds_var.shape!r}")
        if bnds_var.dtype != expected_dtype:
            report.append(
                f"type of bounds coordinate variable {bnds_name!r}"
                f" must be {expected_dtype!r} but was {bnds_var.dtype!r}")
        return None

    return var


def _check_lon_or_lat(dataset, var_name, min_value, max_value, report):
    var = _check_coord_var(dataset, var_name, report)
    if var is None:
        return

    if not np.all(np.isfinite(var)):
        report.append(f"values of coordinate variable {var_name!r} must be finite")

    if np.min(var) < min_value or np.max(var) > max_value:
        report.append(f"values of coordinate variable {var_name!r}"
                      f" must be in the range {min_value} to {max_value}")


def _check_time(dataset, name, report):
    var = _check_coord_var(dataset, name, report)
    if var is None:
        return

    if not np.issubdtype(var.dtype, np.datetime64):
        report.append(f"type of coordinate variable {name!r} must be datetime64")

    if not np.all(np.diff(var.astype(np.float64)) > 0):
        report.append(f"values of coordinate variable {name!r} must be monotonic increasing")



# The MIT License (MIT)
# Copyright (c) 2022 by the xcube team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

from typing import Any, Dict, Union, Optional

import shapely.geometry
import shapely.wkt
from shapely.errors import WKTReadingError

from xcube.constants import LOG
from xcube.core.geom import get_box_split_bounds_geometry
from xcube.util.perf import measure_time
from .context import ALL_PLACES
from .context import PlacesContext
from ...server.api import ApiError
from ...util.assertions import assert_instance

GeoJsonFeatureCollection = Dict
GeoJsonFeature = Dict

GeometryLike = Union[str, Dict, shapely.geometry.base.BaseGeometry]


def find_places(
        ctx: PlacesContext,
        place_group_id: str,
        base_url: str,
        query_geometry: Optional[GeometryLike] = None,
        query_expr: Optional[Any] = None,
        comb_op: str = "and"
) -> GeoJsonFeatureCollection:
    try:
        if isinstance(query_geometry, str):
            try:
                query_geometry = get_box_split_bounds_geometry(
                    *[float(s) for s in query_geometry.split(",")]
                )
            except (TypeError, ValueError) as e:
                query_geometry = shapely.wkt.loads(query_geometry)
        elif isinstance(query_geometry, dict):
            if query_geometry["type"] == "FeatureCollection":
                query_geometry = shapely.geometry.shape(
                    query_geometry["features"][0]["geometry"]
                )
            elif query_geometry["type"] == "Feature":
                query_geometry = shapely.geometry.shape(
                    query_geometry["geometry"]
                )
            else:
                query_geometry = shapely.geometry.shape(
                    query_geometry
                )
        elif query_geometry is not None:
            assert_instance(query_geometry,
                            shapely.geometry.base.BaseGeometry,
                            name='query_geometry')
    except (WKTReadingError, TypeError, IndexError, ValueError, KeyError) as e:
        raise ApiError.BadRequest(
            "Received invalid geometry bbox, geometry WKT, or GeoJSON object"
        ) from e

    with measure_time() as cm:
        features = _find_places(ctx,
                                place_group_id,
                                base_url,
                                query_geometry,
                                query_expr,
                                comb_op)
    LOG.info(f"{len(features)} places found within {cm.duration} seconds")
    return features


def _find_places(
        ctx: PlacesContext,
        place_group_id: str,
        base_url: str,
        query_geometry: Optional[shapely.geometry.base.BaseGeometry] = None,
        query_expr: Optional[Any] = None,
        comb_op: str = "and"
) -> GeoJsonFeatureCollection:
    if comb_op is not None and comb_op != "and":
        raise ApiError.NotImplemented("comb_op not yet supported")

    if place_group_id == ALL_PLACES:
        place_groups = ctx.get_global_place_groups(base_url,
                                                   load_features=True)
        features = []
        for place_group in place_groups:
            features.extend(place_group['features'])
        feature_collection = dict(type="FeatureCollection",
                                  features=features)
    else:
        feature_collection = ctx.get_global_place_group(place_group_id,
                                                        base_url,
                                                        load_features=True)
        feature_collection = dict(type="FeatureCollection",
                                  features=feature_collection['features'])

    if query_geometry is None:
        if query_expr is None:
            return feature_collection
        else:
            raise ApiError.NotImplemented("query_expr not yet supported")
    else:
        matching_places = []
        if query_expr is None:
            for feature in feature_collection['features']:
                geometry = shapely.geometry.shape(feature['geometry'])
                if geometry.intersects(query_geometry):
                    matching_places.append(feature)
        else:
            raise NotImplementedError()
        return dict(type="FeatureCollection",
                    features=matching_places)



# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Sequence

from xcube.core.store import DATASET_TYPE
from xcube.core.store import DataStoreError
from xcube.core.store import DataStorePool
from xcube.core.store import DatasetDescriptor
from xcube.core.store import get_data_store_instance
from xcube.core.store import new_data_opener
from xcube.util.assertions import assert_instance
from xcube.util.assertions import assert_true
from xcube.util.progress import observe_progress
from ..config import InputConfig
from ..error import CubeGeneratorError


class DatasetsDescriber:

    def __init__(self,
                 input_configs: Sequence[InputConfig],
                 store_pool: DataStorePool = None):
        assert_true(len(input_configs) > 0,
                    'At least one input must be given')
        if store_pool is not None:
            assert_instance(store_pool, DataStorePool, 'store_pool')
        self._input_configs = input_configs
        self._store_pool = store_pool

    def describe_datasets(self) -> Sequence[DatasetDescriptor]:
        descriptors = []
        with observe_progress('Fetching dataset information',
                              len(self._input_configs)) as progress:
            for input_config in self._input_configs:
                descriptors.append(self._describe_dataset(input_config))
                progress.worked(1)
        return descriptors

    def _describe_dataset(self, input_config: InputConfig) \
            -> DatasetDescriptor:
        opener_id = input_config.opener_id
        store_params = input_config.store_params or {}
        if input_config.store_id:
            store_instance = get_data_store_instance(
                input_config.store_id,
                store_params=store_params,
                store_pool=self._store_pool
            )
            opener = store_instance.store
        else:
            opener = new_data_opener(opener_id)
        try:
            descriptor = opener.describe_data(input_config.data_id,
                                              data_type=DATASET_TYPE)
        except DataStoreError as dse:
            raise CubeGeneratorError(f'{dse}',
                                     status_code=400) from dse
        if not isinstance(descriptor, DatasetDescriptor):
            raise RuntimeError(f'internal error: data store '
                               f'"{input_config.store_id}": '
                               f'expected DatasetDescriptor but got '
                               f'a {type(descriptor)}')
        return descriptor
